# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

import addonmanager_freecad_interface as fci

# Unit tests for the Addon Manager module
from AddonManagerTest.app.test_utilities import (
    TestUtilities as AddonManagerTestUtilities,
)
from AddonManagerTest.app.test_addon import (
    TestAddon as AddonManagerTestAddon,
)
from AddonManagerTest.app.test_macro import (
    TestMacro as AddonManagerTestMacro,
)
from AddonManagerTest.app.test_git import (
    TestGit as AddonManagerTestGit,
)
from AddonManagerTest.app.test_installer import (
    TestAddonInstaller as AddonManagerTestAddonInstaller,
    TestMacroInstaller as AddonManagerTestMacroInstaller,
)
from AddonManagerTest.app.test_dependency_installer import (
    TestDependencyInstaller as AddonManagerTestDependencyInstaller,
)
from AddonManagerTest.app.test_uninstaller import (
    TestAddonUninstaller as AddonManagerTestAddonUninstaller,
    TestMacroUninstaller as AddonManagerTestMacroUninstaller,
)
from AddonManagerTest.app.test_freecad_interface import (
    TestConsole as AddonManagerTestConsole,
    TestParameters as AddonManagerTestParameters,
    TestDataPaths as AddonManagerTestDataPaths,
)
from AddonManagerTest.app.test_metadata import (
    TestDependencyType as AddonManagerTestDependencyType,
    TestMetadataReader as AddonManagerTestMetadataReader,
    TestMetadataReaderIntegration as AddonManagerTestMetadataReaderIntegration,
    TestUrlType as AddonManagerTestUrlType,
    TestVersion as AddonManagerTestVersion,
    TestMetadataAuxiliaryFunctions as AddonManagerTestMetadataAuxiliaryFunctions,
)


class TestListTerminator:
    pass


# Basic usage mostly to get static analyzers to stop complaining about unused imports
loaded_gui_tests = [
    AddonManagerTestUtilities,
    AddonManagerTestAddon,
    AddonManagerTestMacro,
    AddonManagerTestGit,
    AddonManagerTestAddonInstaller,
    AddonManagerTestMacroInstaller,
    AddonManagerTestDependencyInstaller,
    AddonManagerTestAddonUninstaller,
    AddonManagerTestMacroUninstaller,
    AddonManagerTestConsole,
    AddonManagerTestParameters,
    AddonManagerTestDataPaths,
    AddonManagerTestDependencyType,
    AddonManagerTestMetadataReader,
    AddonManagerTestMetadataReaderIntegration,
    AddonManagerTestUrlType,
    AddonManagerTestVersion,
    AddonManagerTestMetadataAuxiliaryFunctions,
    TestListTerminator,  # Needed to prevent the last test from running twice
]
for test in loaded_gui_tests:
    fci.Console.PrintLog(f"Loaded tests from {test.__name__}\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the viewprovider code for the Dimension objects.

These include linear dimensions, including radius and diameter,
as well as angular dimensions.
They inherit their behavior from the base Annotation viewprovider.
"""
## @package view_dimension
# \ingroup draftviewproviders
# \brief Provides the viewprovider code for the Dimension objects.

import math
import pivy.coin as coin
import lazy_loader.lazy_loader as lz
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import DraftVecUtils
from draftutils import gui_utils
from draftutils import params
from draftutils import units
from draftutils import utils
from draftviewproviders.view_draft_annotation import ViewProviderDraftAnnotation

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")
DraftGeomUtils = lz.LazyLoader("DraftGeomUtils", globals(), "DraftGeomUtils")

## \addtogroup draftviewproviders
# @{


class ViewProviderDimensionBase(ViewProviderDraftAnnotation):
    """The base viewprovider for the Draft Dimensions object.

    This class is not used directly, but inherited by dimension
    viewproviders like linear, radial, and angular.

    Dimension nomeclature
    ---------------------
    The dimension object depends on various variables to draw the lines
    that are drawn on 3D view.
    ::
            |              txt               |      e
        ----b---------a----------------------b----
            |                                |
            |                                |      d
            |                                |

        c  (t)                              (t)   c

    From the object class

    * `a`, `Dimline`, point through which the dimension line goes through

    From the viewprovider class

    * `b`, `ArrowTypeStart`, `ArrowTypeEnd`, `ArrowSizeStart`, `ArrowSizeEnd`, the symbol shown on the endpoints
    * `c`, `DimOvershoot`, extension to the dimension line going through `a`
    * `d`, `ExtLines`, distance to target `(t)`
    * `e`, `ExtOvershoot`, extension in the opposite direction to `(t)`
    * `txt`, text label showing the value of the measurement

    Coin object structure
    ---------------------
    The scenegraph is set from two main nodes.
    ::
        vobj.node_wld.linecolor
                     .drawstyle
                     .lineswitch_wld.coords
                                    .line
                                    .marks
                                    .marksDimOvershoot
                                    .marksExtOvershoot
                     .label_wld.textpos
                               .textcolor
                               .font
                               .text_wld

        vobj.node_scr.linecolor
                     .drawstyle
                     .lineswitch_scr.coords
                                    .line
                                    .marks
                                    .marksDimOvershoot
                                    .marksExtOvershoot
                     .label_scr.textpos
                               .textcolor
                               .font
                               .text_scr
    """

    def set_text_properties(self, vobj, properties):
        """Set text properties only if they don't already exist."""
        super().set_text_properties(vobj, properties)

        if "TextSpacing" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Spacing between text and dimension line")
            vobj.addProperty("App::PropertyLength", "TextSpacing", "Text", _tip, locked=True)
            vobj.TextSpacing = params.get_param("dimspacing")

        if "FlipText" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Rotate the dimension text 180 degrees")
            vobj.addProperty("App::PropertyBool", "FlipText", "Text", _tip, locked=True)
            vobj.FlipText = False

        if "TextPosition" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Text Position.\n" "Leave '(0,0,0)' for automatic position"
            )
            vobj.addProperty(
                "App::PropertyVectorDistance", "TextPosition", "Text", _tip, locked=True
            )
            vobj.TextPosition = App.Vector(0, 0, 0)

        if "Override" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Text override.\n"
                "Write '$dim' so that it is replaced by "
                "the dimension length.",
            )
            vobj.addProperty("App::PropertyString", "Override", "Text", _tip, locked=True)
            vobj.Override = ""

    def set_units_properties(self, vobj, properties):
        """Set unit properties only if they don't already exist."""
        super().set_units_properties(vobj, properties)

        if "Decimals" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "The number of decimals to show")
            vobj.addProperty("App::PropertyInteger", "Decimals", "Units", _tip, locked=True)
            vobj.Decimals = params.get_param("dimPrecision")

        if "ShowUnit" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Show the unit suffix")
            vobj.addProperty("App::PropertyBool", "ShowUnit", "Units", _tip, locked=True)
            vobj.ShowUnit = params.get_param("showUnit")

        if "UnitOverride" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "A unit to express the measurement.\n"
                "Leave blank for system default.\n"
                "Use 'arch' to force US arch notation",
            )
            vobj.addProperty("App::PropertyString", "UnitOverride", "Units", _tip, locked=True)
            vobj.UnitOverride = params.get_param("overrideUnit")

    def set_graphics_properties(self, vobj, properties):
        """Set graphics properties only if they don't already exist."""
        super().set_graphics_properties(vobj, properties)

        if "FlipArrows" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Rotate the dimension arrows 180 degrees")
            vobj.addProperty("App::PropertyBool", "FlipArrows", "Graphics", _tip, locked=True)
            vobj.FlipArrows = False

        if "DimOvershoot" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The distance the dimension line " "is extended\n" "past the extension lines",
            )
            vobj.addProperty("App::PropertyDistance", "DimOvershoot", "Graphics", _tip, locked=True)
            vobj.DimOvershoot = params.get_param("dimovershoot")

        if "ExtLines" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Length of the extension lines")
            vobj.addProperty("App::PropertyDistance", "ExtLines", "Graphics", _tip, locked=True)
            vobj.ExtLines = params.get_param("extlines")

        if "ExtOvershoot" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Length of the extension line\n" "beyond the dimension line"
            )
            vobj.addProperty("App::PropertyDistance", "ExtOvershoot", "Graphics", _tip, locked=True)
            vobj.ExtOvershoot = params.get_param("extovershoot")

        if "ShowLine" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Shows the dimension line and arrows")
            vobj.addProperty("App::PropertyBool", "ShowLine", "Graphics", _tip, locked=True)
            vobj.ShowLine = params.get_param("DimShowLine")

    def getIcon(self):
        """Return the path to the icon used by the viewprovider."""
        return ":/icons/Draft_Dimension_Tree.svg"


class ViewProviderLinearDimension(ViewProviderDimensionBase):
    """The viewprovider for the Linear Dimension objects.

    This includes straight edge measurement, as well as measurement
    of circular edges, and circumferences.
    """

    def set_graphics_properties(self, vobj, properties):
        """Set graphics properties only if they don't already exist."""
        super().set_graphics_properties(vobj, properties)

        self.Object = vobj.Object
        if vobj.Object.Diameter or not self.is_linked_to_circle():
            vobj.ArrowTypeStart = params.get_param("dimsymbolstart")
        else:
            vobj.ArrowTypeStart = "None"
        vobj.ArrowTypeEnd = params.get_param("dimsymbolend")

    def attach(self, vobj):
        """Set up the scene sub-graph of the viewprovider."""
        self.Object = vobj.Object

        self.textpos = coin.SoTransform()
        self.textcolor = coin.SoBaseColor()
        self.font = coin.SoFont()
        self.text_wld = coin.SoAsciiText()  # World orientation. Can be oriented in 3D space.
        self.text_scr = coin.SoText2()  # Screen orientation. Always faces the camera.

        # The text string needs to be initialized to something,
        # otherwise it may cause a crash of the system
        self.text_wld.string = "d"
        self.text_scr.string = "d"
        self.text_wld.justification = coin.SoAsciiText.CENTER
        self.text_scr.justification = coin.SoAsciiText.CENTER

        label_wld = coin.SoSeparator()
        label_wld.addChild(self.textpos)
        label_wld.addChild(self.textcolor)
        label_wld.addChild(self.font)
        label_wld.addChild(self.text_wld)

        label_scr = coin.SoSeparator()
        label_scr.addChild(self.textpos)
        label_scr.addChild(self.textcolor)
        label_scr.addChild(self.font)
        label_scr.addChild(self.text_scr)

        self.coord1 = coin.SoCoordinate3()
        self.trans1 = coin.SoTransform()
        self.coord2 = coin.SoCoordinate3()
        self.trans2 = coin.SoTransform()
        self.transDimOvershoot1 = coin.SoTransform()
        self.transDimOvershoot2 = coin.SoTransform()
        self.transExtOvershoot1 = coin.SoTransform()
        self.transExtOvershoot2 = coin.SoTransform()

        self.linecolor = coin.SoBaseColor()
        self.drawstyle = coin.SoDrawStyle()
        self.coords = coin.SoCoordinate3()
        import PartGui  # Required for "SoBrepEdgeSet" (because a dimension is not a Part::FeaturePython object).

        self.line = coin.SoType.fromName("SoBrepEdgeSet").createInstance()
        self.marks = coin.SoSeparator()
        self.marksDimOvershoot = coin.SoSeparator()
        self.marksExtOvershoot = coin.SoSeparator()

        self.node_wld = coin.SoGroup()
        self.node_wld.addChild(self.linecolor)
        self.node_wld.addChild(self.drawstyle)
        self.lineswitch_wld = coin.SoSwitch()
        self.lineswitch_wld.whichChild = -3
        self.node_wld.addChild(self.lineswitch_wld)
        self.lineswitch_wld.addChild(self.coords)
        self.lineswitch_wld.addChild(self.line)
        self.lineswitch_wld.addChild(self.marks)
        self.lineswitch_wld.addChild(self.marksDimOvershoot)
        self.lineswitch_wld.addChild(self.marksExtOvershoot)
        self.node_wld.addChild(label_wld)

        self.node_scr = coin.SoGroup()
        self.node_scr.addChild(self.linecolor)
        self.node_scr.addChild(self.drawstyle)
        self.lineswitch_scr = coin.SoSwitch()
        self.lineswitch_scr.whichChild = -3
        self.node_scr.addChild(self.lineswitch_scr)
        self.lineswitch_scr.addChild(self.coords)
        self.lineswitch_scr.addChild(self.line)
        self.lineswitch_scr.addChild(self.marks)
        self.lineswitch_scr.addChild(self.marksDimOvershoot)
        self.lineswitch_scr.addChild(self.marksExtOvershoot)
        self.node_scr.addChild(label_scr)

        vobj.addDisplayMode(self.node_wld, "World")
        vobj.addDisplayMode(self.node_scr, "Screen")
        self.updateData(vobj.Object, "Start")
        self.onChanged(vobj, "FontSize")
        self.onChanged(vobj, "FontName")
        self.onChanged(vobj, "TextColor")
        self.onChanged(vobj, "ArrowTypeStart")
        self.onChanged(vobj, "ArrowTypeEnd")
        self.onChanged(vobj, "LineColor")
        self.onChanged(vobj, "DimOvershoot")
        self.onChanged(vobj, "ExtOvershoot")
        self.onChanged(vobj, "ShowLine")
        self.onChanged(vobj, "LineWidth")

    def updateData(self, obj, prop):
        """Execute when a property from the Proxy class is changed.

        It only runs if `Start`, `End`, `Dimline`, or `Direction` changed.
        """
        if prop not in ("Start", "End", "Dimline", "Direction", "Diameter"):
            return

        if obj.Start == obj.End:
            return

        if not hasattr(self, "node_wld"):
            return

        vobj = obj.ViewObject

        if prop == "Diameter":
            if getattr(vobj, "Override", False):
                if obj.Diameter:
                    vobj.Override = vobj.Override.replace("R $dim", "Ø $dim")
                else:
                    vobj.Override = vobj.Override.replace("Ø $dim", "R $dim")

            self.onChanged(vobj, "ArrowTypeStart")
            self.onChanged(vobj, "ArrowTypeEnd")
            return

        # Calculate the 4 points
        #
        #       |        d          |
        #   ---p2-------------c----p3----    c
        #       |                   |
        #       |                   |
        #      p1                  p4
        #
        # - `c` is the `Dimline`, a point that lies on the dimension line
        #   or on its extension.
        # - The line itself between `p2` to `p3` is the `base`.
        # - The distance between `p2` (`base`) to `p1` is `proj`, an extension
        #   line from the dimension to the measured object.
        # - If the `proj` distance is zero, `p1` and `p2` are the same point,
        #   and same with `p3` and `p4`.
        #
        self.p1 = obj.Start
        self.p4 = obj.End
        base = None

        if hasattr(obj, "Direction") and not DraftVecUtils.isNull(obj.Direction):
            v2 = self.p1 - obj.Dimline
            v3 = self.p4 - obj.Dimline
            v2 = DraftVecUtils.project(v2, obj.Direction)
            v3 = DraftVecUtils.project(v3, obj.Direction)
            self.p2 = obj.Dimline + v2
            self.p3 = obj.Dimline + v3
            if DraftVecUtils.equals(self.p2, self.p3):
                base = None
                proj = None
            else:
                base = Part.LineSegment(self.p2, self.p3).toShape()
                proj = DraftGeomUtils.findDistance(self.p1, base)
                if proj:
                    proj = proj.negative()

        if not base:
            if DraftVecUtils.equals(self.p1, self.p4):
                base = None
                proj = None
            else:
                base = Part.LineSegment(self.p1, self.p4).toShape()
                proj = DraftGeomUtils.findDistance(obj.Dimline, base)

            if proj:
                self.p2 = self.p1 + proj.negative()
                self.p3 = self.p4 + proj.negative()
            else:
                self.p2 = self.p1
                self.p3 = self.p4

        if proj:
            if hasattr(vobj, "ExtLines") and hasattr(vobj, "ScaleMultiplier"):
                # The scale multiplier also affects the value
                # of the extension line; this makes sure a maximum length
                # is used if the calculated value is larger than it.
                dmax = vobj.ExtLines.Value * vobj.ScaleMultiplier
                if dmax and proj.Length > dmax:
                    if dmax > 0:
                        self.p1 = self.p2 + DraftVecUtils.scaleTo(proj, dmax)
                        self.p4 = self.p3 + DraftVecUtils.scaleTo(proj, dmax)
                    else:
                        rest = proj.Length + dmax
                        self.p1 = self.p2 + DraftVecUtils.scaleTo(proj, rest)
                        self.p4 = self.p3 + DraftVecUtils.scaleTo(proj, rest)
        else:
            proj = (self.p3 - self.p2).cross(App.Vector(0, 0, 1))

        # Calculate the arrow positions
        p2 = (self.p2.x, self.p2.y, self.p2.z)
        p3 = (self.p3.x, self.p3.y, self.p3.z)

        self.trans1.translation.setValue(p2)
        self.coord1.point.setValue(p2)
        self.trans2.translation.setValue(p3)
        self.coord2.point.setValue(p3)

        # Calculate dimension and extension lines overshoots positions
        self.transDimOvershoot1.translation.setValue(p2)
        self.transDimOvershoot2.translation.setValue(p3)
        self.transExtOvershoot1.translation.setValue(p2)
        self.transExtOvershoot2.translation.setValue(p3)

        # Determine the orientation of the text by using a normal direction.
        # By default the value of +Z will be used, or a calculated value
        # from p2 and p3. So the text will lie on the XY plane
        # or a plane coplanar with p2 and p3.
        u = self.p3 - self.p2
        u.normalize()

        if proj:
            _norm = u.cross(proj)
            norm = _norm.negative()
        else:
            norm = App.Vector(0, 0, 1)

        # If `Normal` exists and is different from the default `(0,0,0)`,
        # it will be used.
        if hasattr(obj, "Normal") and not DraftVecUtils.isNull(obj.Normal):
            norm = App.Vector(obj.Normal)

        if not DraftVecUtils.isNull(norm):
            norm.normalize()

        # Calculate the position of the arrows and extension lines
        v1 = norm.cross(u)
        _plane_rot = DraftVecUtils.getPlaneRotation(u, v1, norm)
        if _plane_rot is not None:
            rot1 = App.Placement(_plane_rot).Rotation.Q
            self.transDimOvershoot1.rotation.setValue((rot1[0], rot1[1], rot1[2], rot1[3]))
            self.transDimOvershoot2.rotation.setValue((rot1[0], rot1[1], rot1[2], rot1[3]))
            self.trot = rot1
        else:
            self.trot = (0, 0, 0, 1)

        if getattr(vobj, "FlipArrows", False):
            u = u.negative()

        v2 = norm.cross(u)
        _plane_rot = DraftVecUtils.getPlaneRotation(u, v2)
        if _plane_rot is not None:
            rot2 = App.Placement(_plane_rot).Rotation.Q
            self.trans1.rotation.setValue((rot2[0], rot2[1], rot2[2], rot2[3]))
            self.trans2.rotation.setValue((rot2[0], rot2[1], rot2[2], rot2[3]))

        if self.p1 != self.p2:
            u3 = self.p1 - self.p2
            u3.normalize()
            v3 = norm.cross(u3)
            _plane_rot = DraftVecUtils.getPlaneRotation(u3, v3)
            if _plane_rot is not None:
                rot3 = App.Placement(_plane_rot).Rotation.Q
                self.transExtOvershoot1.rotation.setValue((rot3[0], rot3[1], rot3[2], rot3[3]))
                self.transExtOvershoot2.rotation.setValue((rot3[0], rot3[1], rot3[2], rot3[3]))

        # Offset is the distance from the dimension line to the textual
        # element that displays the value of the measurement
        if hasattr(vobj, "TextSpacing") and hasattr(vobj, "ScaleMultiplier"):
            ts = vobj.TextSpacing.Value * vobj.ScaleMultiplier
            offset = DraftVecUtils.scaleTo(v1, ts)
        else:
            offset = DraftVecUtils.scaleTo(v1, 0.05)

        if getattr(vobj, "FlipText", False):
            _rott = App.Rotation(self.trot[0], self.trot[1], self.trot[2], self.trot[3])
            self.trot = _rott.multiply(App.Rotation(App.Vector(0, 0, 1), 180)).Q
            offset = offset.negative()

        # On first run the `DisplayMode` enumeration is not set, so we trap
        # the exception and set the display mode using the value
        # in the parameter database
        try:
            m = vobj.DisplayMode
        except AssertionError:
            m = ["World", "Screen"][params.get_param("DefaultAnnoDisplayMode")]

        if m == "Screen":
            offset = offset.negative()

        # The position of the text element in the dimension is provided
        # in absolute coordinates by the value of `TextPosition`,
        # if it is different from the default `(0,0,0)`
        if hasattr(vobj, "TextPosition") and not DraftVecUtils.isNull(vobj.TextPosition):
            self.tbase = vobj.TextPosition
        else:
            # Otherwise the position is calculated from the end points
            # of the dimension line, and the offset that depends
            # on `TextSpacing`
            center = self.p2 + (self.p3 - self.p2).multiply(0.5)
            self.tbase = center + offset

        self.textpos.translation.setValue([self.tbase.x, self.tbase.y, self.tbase.z])
        self.textpos.rotation = coin.SbRotation(
            self.trot[0], self.trot[1], self.trot[2], self.trot[3]
        )

        show_unit = True
        if hasattr(vobj, "ShowUnit"):
            show_unit = vobj.ShowUnit

        # Set text element showing the value of the dimension
        length = (self.p3 - self.p2).Length
        unit = None

        if hasattr(vobj, "UnitOverride"):
            unit = vobj.UnitOverride

        # Special representation if we use 'Building US' scheme
        doc = obj.Document
        if (not unit and doc.UnitSystem == doc.getEnumerationsOfProperty("UnitSystem")[5]) or (
            unit == "arch"
        ):
            self.string = App.Units.Quantity(length, App.Units.Length).UserString
            if self.string.count('"') > 1:
                # multiple inch tokens
                self.string = self.string.replace('"', "", self.string.count('"') - 1)
            sep = params.get_param("FeetSeparator")
            # use a custom separator
            self.string = self.string.replace("' ", "'" + sep)
            self.string = self.string.replace("+", " ")
            self.string = self.string.replace("   ", " ")
            self.string = self.string.replace("  ", " ")
        elif hasattr(vobj, "Decimals"):
            self.string = units.display_external(length, vobj.Decimals, "Length", show_unit, unit)
        else:
            self.string = units.display_external(length, None, "Length", show_unit, unit)

        if getattr(vobj, "Override", False):
            self.string = vobj.Override.replace("$dim", self.string)

        self.text_wld.string = utils.string_encode_coin(self.string)
        self.text_scr.string = utils.string_encode_coin(self.string)

        # Set the lines
        if m == "Screen":
            # Calculate the spacing of the text
            textsize = len(self.string) * vobj.FontSize.Value / 4.0
            spacing = (self.p3 - self.p2).Length / 2.0 - textsize

            self.p2a = self.p2 + DraftVecUtils.scaleTo(self.p3 - self.p2, spacing)
            self.p2b = self.p3 + DraftVecUtils.scaleTo(self.p2 - self.p3, spacing)
            # fmt: off
            self.coords.point.setValues([[self.p1.x, self.p1.y, self.p1.z],
                                         [self.p2.x, self.p2.y, self.p2.z],
                                         [self.p2a.x, self.p2a.y, self.p2a.z],
                                         [self.p2b.x, self.p2b.y, self.p2b.z],
                                         [self.p3.x, self.p3.y, self.p3.z],
                                         [self.p4.x, self.p4.y, self.p4.z]])
            # fmt: on
            # self.line.numVertices.setValues([3, 3])
            self.line.coordIndex.setValues(0, 7, (0, 1, 2, -1, 3, 4, 5))
        else:
            # fmt: off
            self.coords.point.setValues([[self.p1.x, self.p1.y, self.p1.z],
                                         [self.p2.x, self.p2.y, self.p2.z],
                                         [self.p3.x, self.p3.y, self.p3.z],
                                         [self.p4.x, self.p4.y, self.p4.z]])
            # fmt: on
            # self.line.numVertices.setValue(4)
            self.line.coordIndex.setValues(0, 4, (0, 1, 2, 3))

    def onChanged(self, vobj, prop):
        """Execute when a view property is changed."""
        super().onChanged(vobj, prop)

        obj = vobj.Object
        properties = vobj.PropertiesList

        if prop == "ScaleMultiplier" and "ScaleMultiplier" in properties:
            # Update all dimension values
            if hasattr(self, "font"):
                self.font.size = vobj.FontSize.Value * vobj.ScaleMultiplier
            if (
                hasattr(self, "node_wld")
                and hasattr(self, "p2")
                and "ArrowSizeStart" in properties
                and "ArrowSizeEnd" in properties
            ):
                self.remove_dim_arrows()
                self.draw_dim_arrows(vobj)
            if "DimOvershoot" in properties:
                self.remove_dim_overshoot()
                self.draw_dim_overshoot(vobj)
            if "ExtOvershoot" in properties:
                self.remove_ext_overshoot()
                self.draw_ext_overshoot(vobj)

            self.updateData(obj, "Start")

        elif prop == "FontSize" and "FontSize" in properties and "ScaleMultiplier" in properties:
            if hasattr(self, "font"):
                self.font.size = vobj.FontSize.Value * vobj.ScaleMultiplier

        elif prop == "FontName" and "FontName" in properties and hasattr(self, "font"):
            self.font.name = str(vobj.FontName)

        elif prop == "TextColor" and "TextColor" in properties and hasattr(self, "textcolor"):
            col = vobj.TextColor
            self.textcolor.rgb.setValue(col[0], col[1], col[2])

        elif prop == "LineColor" and "LineColor" in properties and hasattr(self, "linecolor"):
            col = vobj.LineColor
            self.linecolor.rgb.setValue(col[0], col[1], col[2])

        elif prop == "LineWidth" and "LineWidth" in properties and hasattr(self, "drawstyle"):
            self.drawstyle.lineWidth = vobj.LineWidth

        elif (
            prop in ("ArrowSizeStart", "ArrowSizeEnd", "ArrowTypeStart", "ArrowTypeEnd")
            and "ArrowSizeStart" in properties
            and "ArrowSizeEnd" in properties
            and "ScaleMultiplier" in properties
            and hasattr(self, "node_wld")
            and hasattr(self, "p2")
        ):
            self.remove_dim_arrows()
            self.draw_dim_arrows(vobj)

        elif (
            prop == "DimOvershoot"
            and "DimOvershoot" in properties
            and "ScaleMultiplier" in properties
        ):
            self.remove_dim_overshoot()
            self.draw_dim_overshoot(vobj)

        elif (
            prop == "ExtOvershoot"
            and "ExtOvershoot" in properties
            and "ScaleMultiplier" in properties
        ):
            self.remove_ext_overshoot()
            self.draw_ext_overshoot(vobj)

        elif prop == "ShowLine" and "ShowLine" in properties:
            if vobj.ShowLine:
                self.lineswitch_wld.whichChild = -3
                self.lineswitch_scr.whichChild = -3
            else:
                self.lineswitch_wld.whichChild = -1
                self.lineswitch_scr.whichChild = -1
        else:
            self.updateData(obj, "Start")

    def remove_dim_arrows(self):
        """Remove dimension arrows in the dimension lines.

        Remove the existing nodes.
        """
        self.node_wld.removeChild(self.marks)
        self.node_scr.removeChild(self.marks)

    def draw_dim_arrows(self, vobj):
        """Draw dimension arrows."""
        if not (
            hasattr(vobj, "ArrowTypeStart")
            and hasattr(vobj, "ArrowTypeEnd")
            and hasattr(vobj, "ArrowSizeStart")
            and hasattr(vobj, "ArrowSizeEnd")
        ):
            return

        # Set scale
        symbol_sta = utils.ARROW_TYPES.index(vobj.ArrowTypeStart)
        symbol_end = utils.ARROW_TYPES.index(vobj.ArrowTypeEnd)
        scale_sta = vobj.ArrowSizeStart.Value * vobj.ScaleMultiplier
        scale_end = vobj.ArrowSizeEnd.Value * vobj.ScaleMultiplier
        self.trans1.scaleFactor.setValue((scale_sta, scale_sta, scale_sta))
        self.trans2.scaleFactor.setValue((scale_end, scale_end, scale_end))

        # Set new nodes
        self.marks = coin.SoSeparator()
        self.marks.addChild(self.linecolor)

        s1 = coin.SoSeparator()
        if symbol_sta == "Circle":
            s1.addChild(self.coord1)
        else:
            s1.addChild(self.trans1)

        s1.addChild(gui_utils.dim_symbol(symbol_sta, invert=False))
        self.marks.addChild(s1)

        s2 = coin.SoSeparator()
        if symbol_end == "Circle":
            s2.addChild(self.coord2)
        else:
            s2.addChild(self.trans2)

        s2.addChild(gui_utils.dim_symbol(symbol_end, invert=True))
        self.marks.addChild(s2)

        self.node_wld.insertChild(self.marks, 2)
        self.node_scr.insertChild(self.marks, 2)

    def remove_dim_overshoot(self):
        """Remove the dimension overshoot lines."""
        self.node_wld.removeChild(self.marksDimOvershoot)
        self.node_scr.removeChild(self.marksDimOvershoot)

    def draw_dim_overshoot(self, vobj):
        """Draw dimension overshoot lines."""
        # Set scale
        s = vobj.DimOvershoot.Value * vobj.ScaleMultiplier
        self.transDimOvershoot1.scaleFactor.setValue((s, s, s))
        self.transDimOvershoot2.scaleFactor.setValue((s, s, s))

        # Remove existing nodes, and set new nodes
        self.marksDimOvershoot = coin.SoSeparator()
        if vobj.DimOvershoot.Value:
            self.marksDimOvershoot.addChild(self.linecolor)

            s1 = coin.SoSeparator()
            s1.addChild(self.transDimOvershoot1)
            s1.addChild(gui_utils.dimDash((-1, 0, 0), (0, 0, 0)))
            self.marksDimOvershoot.addChild(s1)

            s2 = coin.SoSeparator()
            s2.addChild(self.transDimOvershoot2)
            s2.addChild(gui_utils.dimDash((0, 0, 0), (1, 0, 0)))
            self.marksDimOvershoot.addChild(s2)

        self.node_wld.insertChild(self.marksDimOvershoot, 2)
        self.node_scr.insertChild(self.marksDimOvershoot, 2)

    def remove_ext_overshoot(self):
        """Remove dimension extension overshoot lines."""
        self.node_wld.removeChild(self.marksExtOvershoot)
        self.node_scr.removeChild(self.marksExtOvershoot)

    def draw_ext_overshoot(self, vobj):
        """Draw dimension extension overshoot lines."""
        # Set scale
        s = vobj.ExtOvershoot.Value * vobj.ScaleMultiplier
        self.transExtOvershoot1.scaleFactor.setValue((s, s, s))
        self.transExtOvershoot2.scaleFactor.setValue((s, s, s))

        # Set new nodes
        self.marksExtOvershoot = coin.SoSeparator()
        if vobj.ExtOvershoot.Value:
            self.marksExtOvershoot.addChild(self.linecolor)
            s1 = coin.SoSeparator()
            s1.addChild(self.transExtOvershoot1)
            s1.addChild(gui_utils.dimDash((0, 0, 0), (-1, 0, 0)))
            self.marksExtOvershoot.addChild(s1)

            s2 = coin.SoSeparator()
            s2.addChild(self.transExtOvershoot2)
            s2.addChild(gui_utils.dimDash((0, 0, 0), (-1, 0, 0)))
            self.marksExtOvershoot.addChild(s2)

        self.node_wld.insertChild(self.marksExtOvershoot, 2)
        self.node_scr.insertChild(self.marksExtOvershoot, 2)

    def is_linked_to_circle(self):
        """Return true if the dimension measures a circular edge."""
        obj = self.Object
        if obj.LinkedGeometry and len(obj.LinkedGeometry) == 1:
            linked_obj = obj.LinkedGeometry[0][0]
            subelements = obj.LinkedGeometry[0][1]
            if len(subelements) == 1 and "Edge" in subelements[0]:
                sub = subelements[0]
                index = int(sub[4:]) - 1
                edge = linked_obj.Shape.Edges[index]
                if DraftGeomUtils.geomType(edge) == "Circle":
                    return True
        return False

    def getIcon(self):
        """Return the path to the icon used by the viewprovider."""
        if self.is_linked_to_circle():
            return ":/icons/Draft_DimensionRadius.svg"
        return ":/icons/Draft_Dimension_Tree.svg"


# Alias for compatibility with v0.18 and earlier
_ViewProviderDimension = ViewProviderLinearDimension


class ViewProviderAngularDimension(ViewProviderDimensionBase):
    """Viewprovider for the Angular dimension object."""

    def set_graphics_properties(self, vobj, properties):
        """Set graphics properties only if they don't already exist."""
        super().set_graphics_properties(vobj, properties)

        vobj.ArrowTypeStart = params.get_param("dimsymbolstart")
        vobj.ArrowTypeEnd = params.get_param("dimsymbolend")

    def attach(self, vobj):
        """Set up the scene sub-graph of the viewprovider."""
        self.Object = vobj.Object

        self.textpos = coin.SoTransform()
        self.textcolor = coin.SoBaseColor()
        self.font = coin.SoFont()
        self.text_wld = coin.SoAsciiText()  # World orientation. Can be oriented in 3D space.
        self.text_scr = coin.SoText2()  # Screen orientation. Always faces the camera.

        # The text string needs to be initialized to something,
        # otherwise it may cause a crash of the system
        self.text_wld.string = "d"
        self.text_scr.string = "d"
        self.text_wld.justification = coin.SoAsciiText.CENTER
        self.text_scr.justification = coin.SoAsciiText.CENTER

        label_wld = coin.SoSeparator()
        label_wld.addChild(self.textpos)
        label_wld.addChild(self.textcolor)
        label_wld.addChild(self.font)
        label_wld.addChild(self.text_wld)

        label_scr = coin.SoSeparator()
        label_scr.addChild(self.textpos)
        label_scr.addChild(self.textcolor)
        label_scr.addChild(self.font)
        label_scr.addChild(self.text_scr)

        self.coord1 = coin.SoCoordinate3()
        self.trans1 = coin.SoTransform()
        self.coord2 = coin.SoCoordinate3()
        self.trans2 = coin.SoTransform()

        self.linecolor = coin.SoBaseColor()
        self.drawstyle = coin.SoDrawStyle()
        self.coords = coin.SoCoordinate3()
        import PartGui  # Required for "SoBrepEdgeSet" (because a dimension is not a Part::FeaturePython object).

        self.arc = coin.SoType.fromName("SoBrepEdgeSet").createInstance()
        self.marks = coin.SoSeparator()

        self.node_wld = coin.SoGroup()
        self.node_wld.addChild(self.linecolor)
        self.node_wld.addChild(self.drawstyle)
        self.node_wld.addChild(self.coords)
        self.node_wld.addChild(self.arc)
        self.node_wld.addChild(self.marks)
        self.node_wld.addChild(label_wld)

        self.node_scr = coin.SoGroup()
        self.node_scr.addChild(self.linecolor)
        self.node_scr.addChild(self.drawstyle)
        self.node_scr.addChild(self.coords)
        self.node_scr.addChild(self.arc)
        self.node_scr.addChild(self.marks)
        self.node_scr.addChild(label_scr)

        vobj.addDisplayMode(self.node_wld, "World")
        vobj.addDisplayMode(self.node_scr, "Screen")
        self.updateData(vobj.Object, None)
        self.onChanged(vobj, "FontSize")
        self.onChanged(vobj, "FontName")
        self.onChanged(vobj, "TextColor")
        self.onChanged(vobj, "ArrowTypeStart")
        self.onChanged(vobj, "ArrowTypeEnd")
        self.onChanged(vobj, "LineColor")

    def updateData(self, obj, prop):
        """Execute when a property from the Proxy class is changed."""
        if not hasattr(self, "arc"):
            return

        arcsegs = 24

        vobj = obj.ViewObject

        # Determine the orientation of the text by using a normal direction.
        # Also calculate the arc data.
        if DraftVecUtils.isNull(obj.Normal):
            norm = App.Vector(0, 0, 1)
        else:
            norm = obj.Normal

        radius = (obj.Dimline - obj.Center).Length
        self.circle = Part.makeCircle(
            radius, obj.Center, norm, obj.FirstAngle.Value, obj.LastAngle.Value
        )
        self.p2 = self.circle.Vertexes[0].Point
        self.p3 = self.circle.Vertexes[-1].Point
        midp = DraftGeomUtils.findMidpoint(self.circle)
        ray = midp - obj.Center

        # Set text value
        if obj.LastAngle.Value > obj.FirstAngle.Value:
            angle = obj.LastAngle.Value - obj.FirstAngle.Value
        else:
            angle = (360 - obj.FirstAngle.Value) + obj.LastAngle.Value

        show_unit = True
        if hasattr(vobj, "ShowUnit"):
            show_unit = vobj.ShowUnit

        if hasattr(vobj, "Decimals"):
            self.string = units.display_external(angle, vobj.Decimals, "Angle", show_unit)
        else:
            self.string = units.display_external(angle, None, "Angle", show_unit)

        if vobj.Override:
            self.string = vobj.Override.replace("$dim", self.string)

        self.text_wld.string = utils.string_encode_coin(self.string)
        self.text_scr.string = utils.string_encode_coin(self.string)

        # On first run the `DisplayMode` enumeration is not set, so we trap
        # the exception and set the display mode using the value
        # in the parameter database
        try:
            m = vobj.DisplayMode
        except AssertionError:
            m = ["World", "Screen"][params.get_param("DefaultAnnoDisplayMode")]

        # Set the arc
        first = self.circle.FirstParameter
        last = self.circle.LastParameter

        if m == "Screen":
            # Calculate the spacing of the text
            spacing = len(self.string) * vobj.FontSize.Value / 8.0
            pts1 = []
            cut = None
            pts2 = []

            for i in range(arcsegs + 1):
                p = self.circle.valueAt(first + (last - first) / arcsegs * i)
                if (p - midp).Length <= spacing:
                    if cut is None:
                        cut = i
                else:
                    if cut is None:
                        pts1.append([p.x, p.y, p.z])
                    else:
                        pts2.append([p.x, p.y, p.z])

            self.coords.point.setValues(pts1 + pts2)

            pts1_num = len(pts1)
            pts2_num = len(pts2)
            i1 = pts1_num
            i2 = i1 + pts2_num

            self.arc.coordIndex.setValues(
                0, pts1_num + pts2_num + 1, list(range(pts1_num)) + [-1] + list(range(i1, i2))
            )

            if pts1_num >= 3 and pts2_num >= 3:
                self.circle1 = Part.Arc(
                    App.Vector(pts1[0][0], pts1[0][1], pts1[0][2]),
                    App.Vector(pts1[1][0], pts1[1][1], pts1[1][2]),
                    App.Vector(pts1[-1][0], pts1[-1][1], pts1[-1][2]),
                ).toShape()
                self.circle2 = Part.Arc(
                    App.Vector(pts2[0][0], pts2[0][1], pts2[0][2]),
                    App.Vector(pts2[1][0], pts2[1][1], pts2[1][2]),
                    App.Vector(pts2[-1][0], pts2[-1][1], pts2[-1][2]),
                ).toShape()
        else:
            pts = []
            for i in range(arcsegs + 1):
                p = self.circle.valueAt(first + (last - first) / arcsegs * i)
                pts.append([p.x, p.y, p.z])

            self.coords.point.setValues(pts)
            self.arc.coordIndex.setValues(0, arcsegs + 1, list(range(arcsegs + 1)))

        # Set the arrow coords and rotation
        p2 = (self.p2.x, self.p2.y, self.p2.z)
        p3 = (self.p3.x, self.p3.y, self.p3.z)

        self.trans1.translation.setValue(p2)
        self.coord1.point.setValue(p2)
        self.trans2.translation.setValue(p3)
        self.coord2.point.setValue(p3)

        # Calculate small chords to make arrows look better
        if (
            hasattr(vobj, "ArrowSizeStart")
            and hasattr(vobj, "ArrowSizeEnd")
            and hasattr(vobj, "ScaleMultiplier")
            and hasattr(vobj, "FlipArrows")
            and vobj.ArrowSizeStart.Value != 0
            and vobj.ArrowSizeEnd.Value != 0
            and vobj.ScaleMultiplier != 0
        ):
            half_arrow_length_sta = 2 * vobj.ArrowSizeStart.Value * vobj.ScaleMultiplier
            arrow_angle_sta = 2 * math.asin(min(1, half_arrow_length_sta / radius))
            half_arrow_length_end = 2 * vobj.ArrowSizeEnd.Value * vobj.ScaleMultiplier
            arrow_angle_end = 2 * math.asin(min(1, half_arrow_length_end / radius))
            if vobj.FlipArrows:
                arrow_angle_sta = -arrow_angle_sta
                arrow_angle_end = -arrow_angle_end

            u1 = (
                self.circle.valueAt(first + arrow_angle_sta) - self.circle.valueAt(first)
            ).normalize()
            u2 = (
                self.circle.valueAt(last) - self.circle.valueAt(last - arrow_angle_end)
            ).normalize()

            w2 = self.circle.Curve.Axis
            w1 = w2.negative()

            v1 = w1.cross(u1)
            v2 = w2.cross(u2)
            _plane_rot_1 = DraftVecUtils.getPlaneRotation(u1, v1)
            _plane_rot_2 = DraftVecUtils.getPlaneRotation(u2, v2)
            q1 = App.Placement(_plane_rot_1).Rotation.Q
            q2 = App.Placement(_plane_rot_2).Rotation.Q

            self.trans1.rotation.setValue((q1[0], q1[1], q1[2], q1[3]))
            self.trans2.rotation.setValue((q2[0], q2[1], q2[2], q2[3]))

        # Set text position and rotation
        self.tbase = midp
        if hasattr(vobj, "TextPosition") and not DraftVecUtils.isNull(vobj.TextPosition):
            self.tbase = vobj.TextPosition

        u3 = ray.cross(norm).normalize()
        v3 = norm.cross(u3)
        _plane_rot_3 = DraftVecUtils.getPlaneRotation(u3, v3)
        r = App.Placement(_plane_rot_3).Rotation
        offset = r.multVec(App.Vector(0, 1, 0))

        if hasattr(vobj, "TextSpacing") and hasattr(vobj, "ScaleMultiplier"):
            ts = vobj.TextSpacing.Value * vobj.ScaleMultiplier
            offset = DraftVecUtils.scaleTo(offset, ts)
        else:
            offset = DraftVecUtils.scaleTo(offset, 0.05)

        if getattr(vobj, "FlipText", False):
            r = r.multiply(App.Rotation(App.Vector(0, 0, 1), 180))
            offset = offset.negative()

        if m == "Screen":
            offset = offset.negative()

        self.tbase = self.tbase.add(offset)
        q = r.Q
        self.textpos.translation.setValue([self.tbase.x, self.tbase.y, self.tbase.z])
        self.textpos.rotation = coin.SbRotation(q[0], q[1], q[2], q[3])

        # Set the angle property
        _round_1 = round(obj.Angle, utils.precision())
        _round_2 = round(angle, utils.precision())
        if _round_1 != _round_2:
            obj.Angle = angle

    def onChanged(self, vobj, prop):
        """Execute when a view property is changed."""
        super().onChanged(vobj, prop)

        obj = vobj.Object
        properties = vobj.PropertiesList

        if "ScaleMultiplier" in properties and vobj.ScaleMultiplier == 0:
            return

        if prop == "ScaleMultiplier" and "ScaleMultiplier" in properties:
            if hasattr(self, "font"):
                self.font.size = vobj.FontSize.Value * vobj.ScaleMultiplier
            if (
                hasattr(self, "node_wld")
                and hasattr(self, "p2")
                and "ArrowSizeStart" in properties
                and "ArrowSizeEnd" in properties
            ):
                self.remove_dim_arrows()
                self.draw_dim_arrows(vobj)

            self.updateData(obj, None)

        elif prop == "FontSize" and "ScaleMultiplier" in properties:
            if hasattr(self, "font"):
                self.font.size = vobj.FontSize.Value * vobj.ScaleMultiplier

        elif prop == "FontName" and hasattr(self, "font"):
            self.font.name = str(vobj.FontName)

        elif prop == "TextColor" and "TextColor" in properties and hasattr(self, "textcolor"):
            col = vobj.TextColor
            self.textcolor.rgb.setValue(col[0], col[1], col[2])

        elif prop == "LineColor" and "LineColor" in properties and hasattr(self, "linecolor"):
            col = vobj.LineColor
            self.linecolor.rgb.setValue(col[0], col[1], col[2])

        elif prop == "LineWidth" and hasattr(self, "drawstyle"):
            self.drawstyle.lineWidth = vobj.LineWidth

        elif (
            prop in ("ArrowSizeStart", "ArrowTypeStart", "ArrowSizeEnd", "ArrowTypeEnd")
            and "ScaleMultiplier" in properties
            and hasattr(self, "node_wld")
            and hasattr(self, "p2")
        ):
            self.updateData(obj, None)
            self.remove_dim_arrows()
            self.draw_dim_arrows(vobj)

        else:
            self.updateData(obj, None)

    def remove_dim_arrows(self):
        """Remove dimension arrows in the dimension lines.

        Remove the existing nodes.
        """
        self.node_wld.removeChild(self.marks)
        self.node_scr.removeChild(self.marks)

    def draw_dim_arrows(self, vobj):
        """Draw dimension arrows."""
        if not (
            hasattr(vobj, "ArrowTypeStart")
            and hasattr(vobj, "ArrowTypeEnd")
            and hasattr(vobj, "ArrowSizeStart")
            and hasattr(vobj, "ArrowSizeEnd")
        ):
            return

        # Set scale
        symbol_sta = utils.ARROW_TYPES.index(vobj.ArrowTypeStart)
        symbol_end = utils.ARROW_TYPES.index(vobj.ArrowTypeEnd)
        scale_sta = vobj.ArrowSizeStart.Value * vobj.ScaleMultiplier
        scale_end = vobj.ArrowSizeEnd.Value * vobj.ScaleMultiplier
        self.trans1.scaleFactor.setValue((scale_sta, scale_sta, scale_sta))
        self.trans2.scaleFactor.setValue((scale_end, scale_end, scale_end))

        # Set new nodes
        self.marks = coin.SoSeparator()
        self.marks.addChild(self.linecolor)

        s1 = coin.SoSeparator()
        if symbol_sta == "Circle":
            s1.addChild(self.coord1)
        else:
            s1.addChild(self.trans1)
        s1.addChild(gui_utils.dim_symbol(symbol_sta, invert=False))
        self.marks.addChild(s1)

        s2 = coin.SoSeparator()
        if symbol_end == "Circle":
            s2.addChild(self.coord2)
        else:
            s2.addChild(self.trans2)
        s2.addChild(gui_utils.dim_symbol(symbol_end, invert=True))
        self.marks.addChild(s2)

        self.node_wld.insertChild(self.marks, 2)
        self.node_scr.insertChild(self.marks, 2)

    def getIcon(self):
        """Return the path to the icon used by the viewprovider."""
        return ":/icons/Draft_DimensionAngular.svg"


# Alias for compatibility with v0.18 and earlier
_ViewProviderAngularDimension = ViewProviderAngularDimension

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later
# /**************************************************************************
#                                                                           *
#    Copyright (c) 2024 Ondsel <development@ondsel.com>                     *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# **************************************************************************/

import math

import FreeCAD as App


if App.GuiUp:
    import FreeCADGui as Gui

__title__ = "Assembly Marker Inventor object"
__author__ = "Ondsel"
__url__ = "https://www.freecad.org"

from pivy import coin
import UtilsAssembly
import Preferences


class SoSwitchMarker(coin.SoSwitch):
    def __init__(self, vobj):
        super().__init__()  # Initialize the SoSwitch base class

        self.axis_thickness = 3
        self.scaleFactor = 40

        view_params = App.ParamGet("User parameter:BaseApp/Preferences/View")
        param_x_axis_color = view_params.GetUnsigned("AxisXColor", 0xCC333300)
        param_y_axis_color = view_params.GetUnsigned("AxisYColor", 0x33CC3300)
        param_z_axis_color = view_params.GetUnsigned("AxisZColor", 0x3333CC00)

        self.x_axis_so_color = coin.SoBaseColor()
        self.x_axis_so_color.rgb.setValue(UtilsAssembly.color_from_unsigned(param_x_axis_color))
        self.y_axis_so_color = coin.SoBaseColor()
        self.y_axis_so_color.rgb.setValue(UtilsAssembly.color_from_unsigned(param_y_axis_color))
        self.z_axis_so_color = coin.SoBaseColor()
        self.z_axis_so_color.rgb.setValue(UtilsAssembly.color_from_unsigned(param_z_axis_color))

        self.app_obj = vobj.Object
        app_doc = self.app_obj.Document
        self.gui_doc = Gui.getDocument(app_doc)

        self.transform = coin.SoTransform()

        self.draw_style = coin.SoDrawStyle()
        self.draw_style.style = coin.SoDrawStyle.LINES
        self.draw_style.lineWidth = self.axis_thickness

        self.pick = coin.SoPickStyle()
        self.setPickableState(True)

        JCS = coin.SoAnnotation()
        JCS.addChild(self.transform)
        JCS.addChild(self.pick)

        base_plane_sep = self.plane_sep(0.4, 15)
        X_axis_sep = self.line_sep([0.5, 0, 0], [1, 0, 0], self.x_axis_so_color)
        Y_axis_sep = self.line_sep([0, 0.5, 0], [0, 1, 0], self.y_axis_so_color)
        Z_axis_sep = self.line_sep([0, 0, 0], [0, 0, 1], self.z_axis_so_color)

        JCS.addChild(base_plane_sep)
        JCS.addChild(X_axis_sep)
        JCS.addChild(Y_axis_sep)
        JCS.addChild(Z_axis_sep)

        switch_JCS = coin.SoSwitch()
        self.addChild(JCS)
        self.whichChild = coin.SO_SWITCH_NONE

    def line_sep(self, startPoint, endPoint, soColor):
        line = coin.SoLineSet()
        line.numVertices.setValue(2)
        coords = coin.SoCoordinate3()
        coords.point.setValues(0, [startPoint, endPoint])

        axis_sep = coin.SoAnnotation()
        axis_sep.addChild(self.draw_style)
        axis_sep.addChild(soColor)
        axis_sep.addChild(coords)
        axis_sep.addChild(line)

        scale = coin.SoType.fromName("SoShapeScale").createInstance()
        scale.setPart("shape", axis_sep)
        scale.scaleFactor = self.scaleFactor

        return scale

    def plane_sep(self, size, num_vertices):
        coords = coin.SoCoordinate3()

        for i in range(num_vertices):
            angle = float(i) / num_vertices * 2.0 * math.pi
            x = math.cos(angle) * size
            y = math.sin(angle) * size
            coords.point.set1Value(i, x, y, 0)

        face = coin.SoFaceSet()
        face.numVertices.setValue(num_vertices)

        transform = coin.SoTransform()
        transform.translation.setValue(0, 0, 0)

        draw_style = coin.SoDrawStyle()
        draw_style.style = coin.SoDrawStyle.FILLED

        material = coin.SoMaterial()
        material.diffuseColor.setValue([0.5, 0.5, 0.5])
        material.ambientColor.setValue([0.5, 0.5, 0.5])
        material.specularColor.setValue([0.5, 0.5, 0.5])
        material.emissiveColor.setValue([0.5, 0.5, 0.5])
        material.transparency.setValue(0.3)

        face_sep = coin.SoAnnotation()
        face_sep.addChild(transform)
        face_sep.addChild(draw_style)
        face_sep.addChild(material)
        face_sep.addChild(coords)
        face_sep.addChild(face)

        scale = coin.SoType.fromName("SoShapeScale").createInstance()
        scale.setPart("shape", face_sep)
        scale.scaleFactor = self.scaleFactor

        return scale

    def set_marker_placement(self, placement, ref):
        # change plc to be relative to the origin of the document.
        global_plc = UtilsAssembly.getGlobalPlacement(ref)
        placement = global_plc * placement

        t = placement.Base
        self.transform.translation.setValue(t.x, t.y, t.z)

        r = placement.Rotation.Q
        self.transform.rotation.setValue(r[0], r[1], r[2], r[3])

    def setPickableState(self, state: bool):
        """Set JCS selectable or unselectable in 3D view"""
        if not state:
            self.pick.style.setValue(coin.SoPickStyle.UNPICKABLE)
        else:
            self.pick.style.setValue(coin.SoPickStyle.SHAPE_ON_TOP)

    def show_marker(self, visible, placement=None, ref=None):
        if visible:
            self.whichChild = coin.SO_SWITCH_ALL
            self.set_marker_placement(placement, ref)
        else:
            self.whichChild = coin.SO_SWITCH_NONE

    def onChanged(self, vp, prop):
        if prop == "color_X_axis":
            c = vp.getPropertyByName("color_X_axis")
            self.x_axis_so_color.rgb.setValue(c[0], c[1], c[2])
        if prop == "color_Y_axis":
            c = vp.getPropertyByName("color_Y_axis")
            self.y_axis_so_color.rgb.setValue(c[0], c[1], c[2])
        if prop == "color_Z_axis":
            c = vp.getPropertyByName("color_Z_axis")
            self.z_axis_so_color.rgb.setValue(c[0], c[1], c[2])



# (c) 2020 Adam Spontarelli <adam@vector-space.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License (LGPL)
#   as published by the Free Software Foundation; either version 2 of
#   the License, or (at your option) any later version.
#   for detail see the LICENCE text file.
#
#   FCGear is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Library General Public License for more details.
#
#   You should have received a copy of the GNU Library General Public
#   License along with FCGear; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307

from math import cos, sin, tan, sqrt, radians, atan, asin, degrees


def CreateSprocket(w, P, N, Dr):
    """
    Create a sprocket

    w is the wirebuilder object (in which the sprocket will be constructed)
    P is the chain pitch
    N is the number of teeth
    Dr is the roller diameter

    Remaining variables can be found in Standard Handbook of Chains
    """
    Ds = 1.005 * Dr + (0.003 * 25.4)
    R = Ds / 2
    M = 0.8 * Dr * cos(radians(35) + radians(60 / N))
    T = 0.8 * Dr * sin(radians(35) + radians(60 / N))
    E = 1.3025 * Dr + (0.0015 * 25.4)
    W = 1.4 * Dr * cos(radians(180 / N))
    V = 1.4 * Dr * sin(radians(180 / N))
    F = Dr * (
        0.8 * cos(radians(18) - radians(56) / N) + 1.4 * cos(radians(17) - radians(64) / N) - 1.3025
    ) - (0.0015 * 25.4)
    PD = P / (sin(radians(180) / N))
    # H = sqrt(F**2 - (1.4 * Dr - P/2)**2)
    # OD = P * (0.6 + 1/tan(radians(180/N)))

    # The sprocket tooth gullet consists of four segments
    x0 = 0
    y0 = PD / 2 - R

    # ---- Segment 1 -----
    alpha = 35 + 60 / N
    x1 = -R * cos(radians(alpha))
    y1 = PD / 2 - R * sin(radians(alpha))

    # ---- Segment 2 -----
    alpha = 35 + 60 / N
    beta = 18 - 56 / N
    x2 = M - E * cos(radians(alpha - beta))
    y2 = T - E * sin(radians(alpha - beta)) + PD / 2

    # # ---- Segment 3 -----
    y2o = y2 - PD / 2
    hyp = sqrt((-W - x2) ** 2 + (-V - y2o) ** 2)
    AP = sqrt(hyp**2 - F**2)
    gamma = atan((y2o + V) / (x2 + W))
    alpha = asin(AP / hyp)
    beta = 180 - (90 - degrees(alpha)) - (90 - degrees(gamma))
    x3o = AP * sin(radians(beta))
    y3o = AP * cos(radians(beta))
    x3 = x2 - x3o
    y3 = y2 + y3o

    # ---- Segment 4 -----
    alpha = 180 / N
    m = -1 / tan(radians(alpha))
    yf = PD / 2 - V
    A = 1 + m**2
    B = 2 * m * yf - 2 * W
    C = W**2 + yf**2 - F**2
    # x4a = (-B - sqrt(B**2 - 4 * A * C)) / (2*A)
    x4b = (-B + sqrt(B**2 - 4 * A * C)) / (2 * A)
    x4 = -x4b
    y4 = m * x4

    p0 = [x0, y0]
    p1 = [x1, y1]
    p2 = [x2, y2]
    p3 = [x3, y3]
    p4 = [x4, y4]
    p5 = [-x1, y1]
    p6 = [-x2, y2]
    p7 = [-x3, y3]
    p8 = [-x4, y4]

    w.move(p4)  # vectors are lists [x,y]
    w.arc(p3, F, 0)
    w.line(p2)
    w.arc(p1, E, 1)
    w.arc(p0, R, 1)

    # ---- Mirror -----
    w.arc(p5, R, 1)
    w.arc(p6, E, 1)
    w.line(p7)
    w.arc(p8, F, 0)

    # ---- Polar Array ----
    alpha = -radians(360 / N)
    for n in range(1, N):
        # falling gullet slope
        w.arc(rotate(p3, alpha * n), F, 0)
        w.line(rotate(p2, alpha * n))
        w.arc(rotate(p1, alpha * n), E, 1)
        w.arc(rotate(p0, alpha * n), R, 1)

        # rising gullet slope
        w.arc(rotate(p5, alpha * n), R, 1)
        w.line(rotate(p6, alpha * n))
        w.arc(rotate(p7, alpha * n), E, 0)
        w.arc(rotate(p8, alpha * n), F, 0)

    w.close()

    return w


def rotate(pt, rads):
    """
    rotate pt by rads radians about origin
    """
    sinA = sin(rads)
    cosA = cos(rads)
    return (pt[0] * cosA - pt[1] * sinA, pt[0] * sinA + pt[1] * cosA)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2015 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


from PySide import QtCore, QtGui
import FreeCAD
import FreeCADGui
import Path
from PySide.QtCore import QT_TRANSLATE_NOOP
import PathScripts.PathUtils as PathUtils

translate = FreeCAD.Qt.translate


class GCodeHighlighter(QtGui.QSyntaxHighlighter):
    def __init__(self, parent=None):
        def convertcolor(c):
            return QtGui.QColor(int((c >> 24) & 0xFF), int((c >> 16) & 0xFF), int((c >> 8) & 0xFF))

        super(GCodeHighlighter, self).__init__(parent)
        p = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Editor")
        colors = []
        c = p.GetUnsigned("Number")
        if c:
            colors.append(convertcolor(c))
        else:
            colors.append(QtCore.Qt.red)
        c = p.GetUnsigned("Keyword")
        if c:
            colors.append(convertcolor(c))
        else:
            colors.append(QtGui.QColor(0, 170, 0))
        c = p.GetUnsigned("Define name")
        if c:
            colors.append(convertcolor(c))
        else:
            colors.append(QtGui.QColor(160, 160, 164))

        self.highlightingRules = []
        numberFormat = QtGui.QTextCharFormat()
        numberFormat.setForeground(colors[0])
        self.highlightingRules.append((QtCore.QRegularExpression("[\\-0-9\\.]"), numberFormat))
        keywordFormat = QtGui.QTextCharFormat()
        keywordFormat.setForeground(colors[1])
        keywordFormat.setFontWeight(QtGui.QFont.Bold)
        keywordPatterns = ["\\bG[0-9]+\\b", "\\bM[0-9]+\\b"]
        self.highlightingRules.extend(
            [(QtCore.QRegularExpression(pattern), keywordFormat) for pattern in keywordPatterns]
        )
        speedFormat = QtGui.QTextCharFormat()
        speedFormat.setFontWeight(QtGui.QFont.Bold)
        speedFormat.setForeground(colors[2])
        self.highlightingRules.append((QtCore.QRegularExpression("\\bF[0-9\\.]+\\b"), speedFormat))

    def highlightBlock(self, text):

        for pattern, fmt in self.highlightingRules:
            expression = QtCore.QRegularExpression(pattern)
            index = expression.match(text)
            while index.hasMatch():
                length = index.capturedLength()
                self.setFormat(index.capturedStart(), length, fmt)
                index = expression.match(text, index.capturedStart() + length)


class GCodeEditorDialog(QtGui.QDialog):
    tool = None

    def __init__(self, PathObj, parent=FreeCADGui.getMainWindow()):
        self.PathObj = PathObj
        if hasattr(PathObj, "ToolController"):
            self.tool = PathObj.ToolController.Tool
        else:
            self.tool = None

        QtGui.QDialog.__init__(self, parent)
        layout = QtGui.QVBoxLayout(self)

        p = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/CAM")
        c = p.GetUnsigned("DefaultHighlightPathColor", 4286382335)
        Q = QtGui.QColor(int((c >> 24) & 0xFF), int((c >> 16) & 0xFF), int((c >> 8) & 0xFF))
        highlightcolor = (
            Q.red() / 255.0,
            Q.green() / 255.0,
            Q.blue() / 255.0,
            Q.alpha() / 255.0,
        )

        self.selectionobj = FreeCAD.ActiveDocument.addObject("Path::Feature", "selection")
        self.selectionobj.ViewObject.LineWidth = 4
        self.selectionobj.ViewObject.NormalColor = highlightcolor

        # nice text editor widget for editing the gcode
        self.editor = QtGui.QTextEdit()
        font = QtGui.QFont()
        p = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Editor")
        font.setFamily(p.GetString("Font", "Courier"))
        font.setFixedPitch(True)
        font.setPointSize(p.GetInt("FontSize", 10))
        self.editor.setFont(font)
        self.editor.setText("G01 X55 Y4.5 F300.0")
        layout.addWidget(self.editor)

        # Note
        lab = QtGui.QLabel()
        lab.setText(
            translate(
                "CAM_Inspect",
                "<b>Note</b>: This dialog shows path commands in FreeCAD base units (mm/s). \n Values will be converted to the desired unit during post processing.",
            )
        )
        lab.setWordWrap(True)
        layout.addWidget(lab)

        # OK and Cancel buttons
        self.buttons = QtGui.QDialogButtonBox(
            QtGui.QDialogButtonBox.Close,
            QtCore.Qt.Horizontal,
            self,
        )

        layout.addWidget(self.buttons)
        self.buttons.rejected.connect(self.reject)
        self.editor.selectionChanged.connect(self.highlightpath)
        self.finished.connect(self.cleanup)

        prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/CAM")
        Xpos = int(prefs.GetString("inspecteditorX", "0"))
        Ypos = int(prefs.GetString("inspecteditorY", "0"))
        height = int(prefs.GetString("inspecteditorH", "500"))
        width = int(prefs.GetString("inspecteditorW", "600"))
        self.move(Xpos, Ypos)
        self.resize(width, height)

    def cleanup(self):
        prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/CAM")
        prefs.SetString("inspecteditorX", str(self.x()))
        prefs.SetString("inspecteditorY", str(self.y()))
        prefs.SetString("inspecteditorW", str(self.width()))
        prefs.SetString("inspecteditorH", str(self.height()))
        FreeCAD.ActiveDocument.removeObject(self.selectionobj.Name)

    def highlightpath(self):
        cursor = self.editor.textCursor()
        sp = cursor.selectionStart()
        ep = cursor.selectionEnd()
        cursor.setPosition(sp)
        startrow = cursor.blockNumber()
        cursor.setPosition(ep)
        endrow = cursor.blockNumber()

        commands = PathUtils.getPathWithPlacement(self.PathObj).Commands

        # Derive the starting position for the first selected command
        prevX = prevY = prevZ = None
        prevcommands = commands[:startrow]
        prevcommands.reverse()
        for c in prevcommands:
            if prevX is None:
                if c.Parameters.get("X") is not None:
                    prevX = c.Parameters.get("X")
            if prevY is None:
                if c.Parameters.get("Y") is not None:
                    prevY = c.Parameters.get("Y")
            if prevZ is None:
                if c.Parameters.get("Z") is not None:
                    prevZ = c.Parameters.get("Z")
            if prevX is not None and prevY is not None and prevZ is not None:
                break
        if prevX is None:
            prevX = 0.0
        if prevY is None:
            prevY = 0.0
        if prevZ is None:
            prevZ = 0.0

        # Build a new path with selection
        p = Path.Path()
        firstrapid = Path.Command("G0", {"X": prevX, "Y": prevY, "Z": prevZ})

        selectionpath = [firstrapid] + commands[startrow : endrow + 1]
        p.Commands = selectionpath
        self.selectionobj.Path = p

        if self.tool is not None:
            self.tool.Placement.Base.x = prevX
            self.tool.Placement.Base.y = prevY
            self.tool.Placement.Base.z = prevZ


def show(obj):
    "show(obj): shows the G-code data of the given Path object in a dialog"

    prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/CAM")
    # default Max Highlighter Size = 512 Ko
    defaultMHS = 512 * 1024
    mhs = prefs.GetUnsigned("inspecteditorMaxHighlighterSize", defaultMHS)

    if hasattr(obj, "Path"):
        if obj.Path:
            dia = GCodeEditorDialog(obj)
            dia.editor.setText(obj.Path.toGCode())
            gcodeSize = len(dia.editor.toPlainText())
            if gcodeSize <= mhs:
                # because of poor performance, syntax highlighting is
                # limited to mhs octets (default 512 KB).
                # It seems than the response time curve has an inflexion near 500 KB
                # beyond 500 KB, the response time increases exponentially.
                dia.highlighter = GCodeHighlighter(dia.editor.document())
            else:
                FreeCAD.Console.PrintMessage(
                    translate(
                        "Path",
                        "GCode size too big ({} o), disabling syntax highlighter.".format(
                            gcodeSize
                        ),
                    )
                )
            result = dia.exec_()
            # exec_() returns 0 or 1 depending on the button pressed (Ok or
            # Cancel)
            if result:
                p = Path.Path(dia.editor.toPlainText())
                FreeCAD.ActiveDocument.openTransaction("Edit Path")
                obj.Path = p
                FreeCAD.ActiveDocument.commitTransaction()
                FreeCAD.ActiveDocument.recompute()


class CommandPathInspect:
    def GetResources(self):
        return {
            "Pixmap": "CAM_Inspect",
            "MenuText": QT_TRANSLATE_NOOP("CAM_Inspect", "Inspect Toolpath"),
            "Accel": "P, I",
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_Inspect", "Inspects the contents of a toolpath object"
            ),
        }

    def IsActive(self):
        selection = FreeCADGui.Selection.getSelection()
        if len(selection) == 0:
            return False
        obj = selection[0]
        return hasattr(obj, "Path") and len(obj.Path.Commands) > 0

    def Activated(self):
        # check that the selection contains exactly what we want
        selection = FreeCADGui.Selection.getSelection()
        if len(selection) != 1:
            FreeCAD.Console.PrintError(
                translate("CAM_Inspect", "Select exactly one path object") + "\n"
            )
            return
        if not (selection[0].isDerivedFrom("Path::Feature")):
            FreeCAD.Console.PrintError(
                translate("CAM_Inspect", "Select exactly one path object") + "\n"
            )
            return

        # if everything is ok, execute
        FreeCADGui.addModule("Path.Main.Gui.Inspect")
        FreeCADGui.doCommand(
            "Path.Main.Gui.Inspect.show(FreeCAD.ActiveDocument." + selection[0].Name + ")"
        )


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_Inspect", CommandPathInspect())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create circular Array objects."""
## @package make_circulararray
# \ingroup draftmake
# \brief Provides functions to create circular Array objects.

## \addtogroup draftmake
# @{
import FreeCAD as App
import draftmake.make_array as make_array
import draftutils.utils as utils

from draftutils.messages import _err
from draftutils.translate import translate


def make_circular_array(
    base_object,
    r_distance=100,
    tan_distance=50,
    number=3,
    symmetry=1,
    axis=App.Vector(0, 0, 1),
    center=App.Vector(0, 0, 0),
    use_link=True,
):
    """Create a circular array from the given object.

    Parameters
    ----------
    base_object: Part::Feature or str
        Any of object that has a `Part::TopoShape` that can be duplicated.
        This means most 2D and 3D objects produced with any workbench.
        If it is a string, it must be the `Label` of that object.
        Since a label is not guaranteed to be unique in a document,
        it will use the first object found with this label.

    r_distance: float, optional
        It defaults to `100`.
        Radial distance to the next ring of circular arrays.

    tan_distance: float, optional
        It defaults to `50`.
        The tangential distance between two elements located
        in the same circular ring.
        The tangential distance together with the radial distance
        determine how many copies are created.

    number: int, optional
        It defaults to 3.
        The number of layers or rings of repeated objects.
        The original object stays at the center, and is counted
        as a layer itself. So, if you want at least one layer of circular
        copies, this number must be at least 2.

    symmetry: int, optional
        It defaults to 1.
        It indicates how many lines of symmetry the entire circular pattern
        has. That is, with 1, the array is symmetric only after a full
        360 degrees rotation.

        When it is 2, the array is symmetric at 0 and 180 degrees.
        When it is 3, the array is symmetric at 0, 120, and 240 degrees.
        When it is 4, the array is symmetric at 0, 90, 180, and 270 degrees.
        Et cetera.

    axis: Base::Vector3, optional
        It defaults to `App.Vector(0, 0, 1)` or the `+Z` axis.
        The unit vector indicating the axis of rotation.

    center: Base::Vector3, optional
        It defaults to `App.Vector(0, 0, 0)` or the global origin.
        The point through which the `axis` passes to define
        the axis of rotation.

    use_link: bool, optional
        It defaults to `True`.
        If it is `True` the produced copies are not `Part::TopoShape` copies,
        but rather `App::Link` objects.
        The Links repeat the shape of the original `base_object` exactly,
        and therefore the resulting array is more memory efficient.

        Also, when `use_link` is `True`, the `Fuse` property
        of the resulting array does not work; the array doesn't
        contain separate shapes, it only has the original shape repeated
        many times, so there is nothing to fuse together.

        If `use_link` is `False` the original shape is copied many times.
        In this case the `Fuse` property is able to fuse
        all copies into a single object, if they touch each other.

    Returns
    -------
    Part::FeaturePython
        A scripted object of type `'Array'`.
        Its `Shape` is a compound of the copies of the original object.

    None
        If there is a problem it will return `None`.

    See Also
    --------
    make_ortho_array, make_polar_array, make_path_array, make_point_array
    """
    _name = "make_circular_array"

    found, base_object = utils.find_object(base_object, doc=App.activeDocument())
    if not found:
        _err(translate("draft", "Wrong input: base_object not in document."))
        return None

    try:
        utils.type_check(
            [
                (r_distance, (int, float, App.Units.Quantity)),
                (tan_distance, (int, float, App.Units.Quantity)),
            ],
            name=_name,
        )
    except TypeError:
        _err(translate("draft", "Wrong input: must be a number or quantity."))
        return None

    try:
        utils.type_check([(number, int), (symmetry, int)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be an integer number."))
        return None

    try:
        utils.type_check([(axis, App.Vector), (center, App.Vector)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a vector."))
        return None

    use_link = bool(use_link)

    new_obj = make_array.make_array(
        base_object,
        arg1=r_distance,
        arg2=tan_distance,
        arg3=axis,
        arg4=center,
        arg5=number,
        arg6=symmetry,
        use_link=use_link,
    )
    return new_obj


## @}



# /***************************************************************************
# *   Copyright (c) 2018 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

# This is a temporary replacement for C++-powered Container class that should be eventually introduced into FreeCAD


class Container(object):
    """Container class: a unified interface for container objects, such as Group, Part, Body, or Document.
    This is a temporary implementation."""

    Object = None  # DocumentObject or Document, the actual container

    def __init__(self, obj):
        self.Object = obj

    def self_check(self):
        if self.Object is None:
            raise ValueError("Null!")
        if not isAContainer(self.Object, links_too=True):
            raise NotAContainerError(self.Object)

    def getAllChildren(self):
        """Returns all objects directly contained by the container. all = static + dynamic."""
        return self.getStaticChildren() + self.getDynamicChildren()

    def getStaticChildren(self):
        """Returns children tightly bound to the container, such as Origin. The key thing
        about them is that they are not supposed to be removed or added from/to the container."""

        self.self_check()
        container = self.Object
        if container.isDerivedFrom("App::Document"):
            return []
        elif container.hasExtension("App::OriginGroupExtension"):
            if container.Origin is not None:
                return [container.Origin]
            else:
                return []
        elif container.isDerivedFrom("App::Origin"):
            return container.OriginFeatures
        elif container.hasExtension("App::GroupExtension"):
            return []
        elif container.hasChildElement():  # Link
            return []
        raise RuntimeError("getStaticChildren: unexpected container type!")

    def getDynamicChildren(self):
        """Returns dynamic children, i.e. the stuff that can be removed from the container."""
        self.self_check()
        container = self.Object

        if container.isDerivedFrom("App::Document"):
            # find all objects not contained by any Part or Body
            result = set(container.Objects)
            for obj in container.Objects:
                if isAContainer(obj):
                    children = set(Container(obj).getAllChildren())
                    result = result - children
            return list(result)
        elif container.hasExtension("App::GroupExtension"):
            result = container.Group
            if container.hasExtension("App::GeoFeatureGroupExtension"):
                # geofeaturegroup's group contains all objects within the CS, we don't want that
                result = [obj for obj in result if obj.getParentGroup() is not container]
            return result
        elif container.isDerivedFrom("App::Origin"):
            return []
        elif container.hasChildElement():
            result = []
            for sub in container.getSubObjects(1):
                sobj = container.getSubObject(sub, retType=1)
                if sobj:
                    result.append(sobj)
            return result

        raise RuntimeError("getDynamicChildren: unexpected container type!")

    def isACS(self):
        """isACS(): returns true if the container forms internal coordinate system."""
        self.self_check()
        container = self.Object

        if container.isDerivedFrom("App::Document"):
            return True  # Document is a special thing... is it a CS or not is a matter of coding convenience.
        elif container.hasExtension("App::GeoFeatureGroupExtension"):
            return True
        elif container.hasChildElement():  # Link
            return True
        else:
            return False

    def isAVisGroup(self):
        """isAVisGroup(): returns True if the container consumes viewproviders of children, and thus affects their visibility."""
        self.self_check()
        container = self.Object

        if container.isDerivedFrom("App::Document"):
            return True  # Document is a special thing... Return value is a matter of coding convenience.
        elif container.hasExtension("App::GeoFeatureGroupExtension"):
            return True
        elif container.isDerivedFrom("App::Origin"):
            return True
        elif container.hasChildElement():  # Link
            return True
        else:
            return False

    def getCSChildren(self):
        if not self.isACS():
            raise TypeError("Container is not a coordinate system")
        container = self.Object
        return _getMetacontainerChildren(self, Container.isACS)

    def getVisGroupChildren(self):
        if not self.isAVisGroup():
            raise TypeError("Container is not a visibility group")
        container = self.Object
        return _getMetacontainerChildren(self, Container.isAVisGroup)

    def isChildVisible(self, obj):
        container = self.Object
        isElementVisible = getattr(container, "isElementVisible", None)
        if not isElementVisible:
            return obj.Visibility
        vis = isElementVisible(obj.Name)
        if vis < 0:
            return obj.Visibility
        return vis > 0

    def hasObject(self, obj):
        """Returns True if the container contains specified object directly."""
        return obj in self.getAllChildren()

    def hasObjectRecursive(self, obj):
        return self.Object in ContainerChain(obj)


def _getMetacontainerChildren(container, isrightcontainer_func):
    """Gathers up children of metacontainer - a container structure formed by containers of specific type.
    For example, coordinate systems form a kind of container structure.

    container: instance of Container class
    isrightcontainer_func: a function f(cnt)->bool, where cnt is a Container object."""

    result = []
    list_traversing_now = [container]  # list of Container instances
    list_to_be_traversed_next = []  # list of Container instances
    visited_containers = set([container.Object])  # set of DocumentObjects

    while len(list_traversing_now) > 0:
        list_to_be_traversed_next = []
        for itcnt in list_traversing_now:
            children = itcnt.getAllChildren()
            result.extend(children)
            for child in children:
                if isAContainer(child):
                    newcnt = Container(child)
                    if not isrightcontainer_func(newcnt):
                        list_to_be_traversed_next.append(newcnt)
        list_traversing_now = list_to_be_traversed_next

    return result


def isAContainer(obj, links_too=False):
    """isAContainer(obj, links_too): returns True if obj is an object container, such as
    Group, Part, Body. The important characteristic of an object being a
    container is that it can be activated to receive new objects. Documents
    are considered containers, too.
    If links_too, App::Link objects are considered containers, too. Then, container tree
    isn't necessarily a tree."""

    if obj.isDerivedFrom("App::Document"):
        return True
    if obj.hasExtension("App::GroupExtension"):
        return True
    if obj.isDerivedFrom("App::Origin"):
        return True
    if obj.hasChildElement():
        return True if links_too else False
    return False


# from Part-o-magic...
def ContainerOf(obj):
    """ContainerOf(obj): returns the container that immediately has obj."""
    cnt = None
    for dep in obj.InList:
        if isAContainer(dep):
            if Container(dep).hasObject(obj):
                if cnt is not None and dep is not cnt:
                    raise ContainerTreeError("Container tree is not a tree")
                cnt = dep
    if cnt is None:
        return obj.Document
    return cnt


def getVisGroupOf(obj):
    chain = VisGroupChain(obj)
    return chain[-1]


# from Part-o-magic... over-engineered, but proven to work
def ContainerChain(feat):
    """ContainerChain(feat): container path to feat (not including feat itself).
    Last container directly contains the feature.
    Example of output:  [<document>,<SuperPart>,<Part>,<Body>]"""

    if feat.isDerivedFrom("App::Document"):
        return []

    list_traversing_now = [feat]
    set_of_deps = set()
    list_of_deps = []

    while len(list_traversing_now) > 0:
        list_to_be_traversed_next = []
        for feat in list_traversing_now:
            for dep in feat.InList:
                if isAContainer(dep) and Container(dep).hasObject(feat):
                    if not (dep in set_of_deps):
                        set_of_deps.add(dep)
                        list_of_deps.append(dep)
                        list_to_be_traversed_next.append(dep)
        if len(list_to_be_traversed_next) > 1:
            raise ContainerTreeError("Container tree is not a tree")
        list_traversing_now = list_to_be_traversed_next

    return [feat.Document] + list_of_deps[::-1]


def CSChain(feat):
    cnt_chain = ContainerChain(feat)
    return [cnt for cnt in cnt_chain if Container(cnt).isACS()]


def VisGroupChain(feat):
    cnt_chain = ContainerChain(feat)
    return [cnt for cnt in cnt_chain if Container(cnt).isAVisGroup()]


class ContainerError(RuntimeError):
    pass


class NotAContainerError(ContainerError):
    def __init__(self, name="None"):
        ContainerError.__init__(self, "'{}' is not recognized as container".format(name))


class ContainerTreeError(ContainerError):
    pass



# SPDX-License-Identifier: LGPL-2.1-or-later
# /**************************************************************************
#                                                                           *
#    Copyright (c) 2024 Ondsel <development@ondsel.com>                     *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# **************************************************************************/

import re
import os
import FreeCAD as App

from PySide.QtCore import QT_TRANSLATE_NOOP

if App.GuiUp:
    import FreeCADGui as Gui
    from PySide import QtCore, QtGui, QtWidgets
    from PySide.QtGui import QIcon
    from PySide.QtCore import QTimer

import UtilsAssembly
import Preferences
import JointObject

translate = App.Qt.translate

__title__ = "Assembly Command Insert New Part"
__author__ = "Ondsel"
__url__ = "https://www.freecad.org"


class CommandInsertNewPart:
    def __init__(self):
        pass

    def GetResources(self):
        return {
            "Pixmap": "Geofeaturegroup",
            "MenuText": QT_TRANSLATE_NOOP("Assembly_InsertNewPart", "New Part"),
            "Accel": "P",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Assembly_InsertNewPart",
                "Insert a new part into the active assembly. The new part's origin can be positioned in the assembly.",
            ),
            "CmdType": "ForEdit",
        }

    def IsActive(self):
        return UtilsAssembly.isAssemblyCommandActive()

    def Activated(self):
        # Check if document is saved before proceeding
        doc = App.ActiveDocument
        if not doc.FileName:
            msgBox = QtWidgets.QMessageBox()
            msgBox.setIcon(QtWidgets.QMessageBox.Warning)
            msgBox.setText(
                translate(
                    "Assembly",
                    "The assembly document must be saved before inserting a new part.",
                )
            )
            msgBox.setWindowTitle(translate("Assembly", "Save Document"))
            saveButton = msgBox.addButton(
                translate("Assembly", "Save"), QtWidgets.QMessageBox.AcceptRole
            )
            msgBox.addButton(QtWidgets.QMessageBox.Cancel)
            msgBox.exec_()
            if msgBox.clickedButton() == saveButton:
                if not Gui.getDocument(doc).saveAs():
                    return
            else:
                return

        panel = TaskAssemblyNewPart()
        dialog = Gui.Control.showDialog(panel)
        if dialog is not None:
            dialog.setAutoCloseOnDeletedDocument(True)
            dialog.setDocumentName(App.ActiveDocument.Name)


class TaskAssemblyNewPart(JointObject.TaskAssemblyCreateJoint):
    def __init__(self):
        super().__init__(0, None, True)

        self.assembly = UtilsAssembly.activeAssembly()

        # Retrieve the existing layout of `self.form`
        mainLayout = self.form.layout()

        # Add a name input
        nameLayout = QtWidgets.QHBoxLayout()
        nameLabel = QtWidgets.QLabel(translate("Assembly", "Part name"))
        self.nameEdit = QtWidgets.QLineEdit()
        nameLayout.addWidget(nameLabel)
        nameLayout.addWidget(self.nameEdit)
        mainLayout.addLayout(nameLayout)
        self.nameEdit.setText(translate("Assembly", "Part"))

        # Add a checkbox
        self.createInNewFileCheck = QtWidgets.QCheckBox(
            translate("Assembly", "Create part in new file")
        )
        mainLayout.addWidget(self.createInNewFileCheck)
        self.createInNewFileCheck.setChecked(
            Preferences.preferences().GetBool("PartInNewFile", True)
        )

        # Wrap the joint creation UI in a groupbox
        jointGroupBox = QtWidgets.QGroupBox(translate("Assembly", "Joint new part origin"))
        jointLayout = QtWidgets.QVBoxLayout(jointGroupBox)
        jointLayout.addWidget(self.jForm)
        jointLayout.setContentsMargins(0, 0, 0, 0)
        jointLayout.setSpacing(0)
        mainLayout.addWidget(jointGroupBox)

        self.link = self.assembly.newObject("App::Link", "Link")
        # add the link as the first object of the joint
        Gui.Selection.addSelection(
            self.assembly.Document.Name, self.assembly.Name, self.link.Name + "."
        )

    def createPart(self):
        partName = self.nameEdit.text()
        newFile = self.createInNewFileCheck.isChecked()

        doc = self.assembly.Document
        if newFile:
            doc = App.newDocument(partName)

        part, body = UtilsAssembly.createPart(partName, doc)

        App.setActiveDocument(self.assembly.Document.Name)

        # Then we need to link the part.
        if newFile:
            # New file must be saved or we can't link
            if not Gui.getDocument(doc).saveAs():
                msgBox = QtWidgets.QMessageBox()
                msgBox.setIcon(QtWidgets.QMessageBox.Warning)
                msgBox.setText(
                    translate(
                        "Assembly",
                        "If the new document is not saved the new part cannot be linked in the assembly.",
                    )
                )
                msgBox.setWindowTitle(translate("Assembly", "Save Document"))
                saveButton = msgBox.addButton(
                    translate("Assembly", "Save"), QtWidgets.QMessageBox.AcceptRole
                )
                cancelButton = msgBox.addButton(
                    translate("Assembly", "Do not Link"), QtWidgets.QMessageBox.RejectRole
                )

                msgBox.exec_()

                if not (msgBox.clickedButton() == saveButton and Gui.getDocument(doc).saveAs()):
                    return

        self.link.LinkedObject = part
        self.link.touch()

        self.link.Label = part.Label

        # Set the body as active in the assembly doc
        self.expandLinkManually(self.link)
        doc = self.assembly.Document
        Gui.getDocument(doc).ActiveView.setActiveObject("pdbody", body)
        doc.recompute()

    def expandLinkManually(self, link):
        # Should not be necessary
        # This is a workaround of https://github.com/FreeCAD/FreeCAD/issues/17904
        mw = Gui.getMainWindow()
        trees = mw.findChildren(QtGui.QTreeWidget)

        Gui.Selection.addSelection(link)
        for tree in trees:
            for item in tree.selectedItems():
                tree.expandItem(item)

    def accept(self):
        if len(self.refs) != 2:
            # if the joint is not complete we cancel the joint but not the new part!
            self.joint.Document.removeObject(self.joint.Name)
        else:
            JointObject.solveIfAllowed(self.assembly)
            self.joint.Visibility = False
            cmds = UtilsAssembly.generatePropertySettings(self.joint)
            Gui.doCommand(cmds)

        self.deactivate()

        self.createPart()

        App.closeActiveTransaction()

        return True

    def deactivate(self):
        Preferences.preferences().SetBool("PartInNewFile", self.createInNewFileCheck.isChecked())
        super().deactivate()


if App.GuiUp:
    Gui.addCommand("Assembly_InsertNewPart", CommandInsertNewPart())



# SPDX-License-Identifier: LGPL-2.1-or-later
# /**************************************************************************
#                                                                           *
#    Copyright (c) 2023 Ondsel <development@ondsel.com>                     *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# **************************************************************************/

import re
import os
import FreeCAD as App

from PySide.QtCore import QT_TRANSLATE_NOOP

if App.GuiUp:
    import FreeCADGui as Gui
    from PySide import QtCore, QtGui, QtWidgets
    from PySide.QtGui import QIcon

import UtilsAssembly
import Preferences
import CommandCreateJoint


__title__ = "Assembly Command Insert Component"
__author__ = "Ondsel"
__url__ = "https://www.freecad.org"


tooltip = QT_TRANSLATE_NOOP(
    "Assembly_InsertLink",
    "<p>Inserts a component into the active assembly. This will create dynamic links to parts, bodies, primitives, and assemblies. To insert external components, make sure that the file is <b>open in the current session</b></p>"
    "<ul>"
    "<li>Insert by left clicking items in the list.</li>"
    "<li>Remove by right clicking items in the list.</li>"
    "<li>Press shift to add several instances of the component while clicking on the view.</li>"
    "</ul>",
)


class CommandGroupInsert:
    def GetCommands(self):
        return ("Assembly_InsertLink", "Assembly_InsertNewPart")

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Assembly_InsertLink",
            "MenuText": QT_TRANSLATE_NOOP("Assembly_Insert", "Insert Component"),
            "ToolTip": tooltip,
            "CmdType": "ForEdit",
        }

    def IsActive(self):
        return UtilsAssembly.isAssemblyCommandActive()


class CommandInsertLink:
    def __init__(self):
        pass

    def GetResources(self):
        return {
            "Pixmap": "Assembly_InsertLink",
            "MenuText": QT_TRANSLATE_NOOP("Assembly_InsertLink", "Component"),
            "Accel": "I",
            "ToolTip": tooltip,
            "CmdType": "ForEdit",
        }

    def IsActive(self):
        return UtilsAssembly.isAssemblyCommandActive()

    def Activated(self):
        assembly = UtilsAssembly.activeAssembly()
        if not assembly:
            return
        view = Gui.activeDocument().activeView()
        self.panel = TaskAssemblyInsertLink(assembly, view)
        Gui.Control.showDialog(self.panel)


class InsertLinkObserver:
    def __init__(self, callback):
        self.callback = callback

    def slotDeletedObject(self, obj):
        self.callback(obj)


class TaskAssemblyInsertLink(QtCore.QObject):
    def __init__(self, assembly, view):
        super().__init__()

        self.assembly = assembly
        self.view = view
        self.doc = App.ActiveDocument
        self.showHidden = False

        self.form = Gui.PySideUic.loadUi(":/panels/TaskAssemblyInsertLink.ui")
        self.form.installEventFilter(self)
        self.form.partList.installEventFilter(self)

        pref = Preferences.preferences()
        self.form.CheckBox_ShowOnlyParts.setChecked(pref.GetBool("InsertShowOnlyParts", False))
        self.form.CheckBox_RigidSubAsm.setChecked(pref.GetBool("InsertRigidSubAssemblies", True))

        # Actions
        self.form.openFileButton.clicked.connect(self.openFiles)
        self.form.partList.itemClicked.connect(self.onItemClicked)
        self.form.filterPartList.textChanged.connect(self.onFilterChange)
        self.form.CheckBox_ShowOnlyParts.stateChanged.connect(self.buildPartList)

        self.form.partList.header().hide()

        self.translation = 0
        # self.partMoving = False
        self.totalTranslation = App.Vector()
        self.prevScreenCenter = App.Vector()
        self.groundedObj = None

        self.insertionStack = []  # used to handle cancellation of insertions.
        self.doc_item_map = {}

        self.buildPartList()

        App.setActiveTransaction("Insert Component")

        # Listen for external deletions to keep the list in sync
        self.docObserver = InsertLinkObserver(self.onObjectDeleted)
        App.addDocumentObserver(self.docObserver)

    def accept(self):
        self.deactivated()

        Gui.addModule("UtilsAssembly")
        commands = "assembly = UtilsAssembly.activeAssembly()\n"
        for insertionItem in self.insertionStack:
            object = insertionItem["addedObject"]
            translation = insertionItem["translation"]

            # Check if object.Name & object.LinkedObject.Name exists
            if (
                not hasattr(object, "Name")
                or not hasattr(object, "LinkedObject")
                or not hasattr(object.LinkedObject, "Name")
            ):
                continue

            commands = commands + (
                f'item = assembly.newObject("App::Link", "{object.Name}")\n'
                f'item.LinkedObject = App.ActiveDocument.getObject("{object.LinkedObject.Name}")\n'
                f'item.Label = "{object.Label}"\n'
            )

            if translation != App.Vector():
                commands = commands + (
                    f"item.Placement.base = App.Vector({translation.x},"
                    f"{translation.y},"
                    f"{translation.z})\n"
                )

        # Ground the first item if that happened
        if self.groundedObj:
            commands = (
                commands
                + f'CommandCreateJoint.createGroundedJoint(App.ActiveDocument.getObject("{self.groundedObj.Name}"))\n'
            )

        Gui.doCommandSkip(commands[:-1])  # Get rid of last \n
        App.closeActiveTransaction()
        return True

    def reject(self):
        self.deactivated()

        App.closeActiveTransaction(True)
        return True

    def deactivated(self):
        if hasattr(self, "docObserver") and self.docObserver:
            App.removeDocumentObserver(self.docObserver)
            self.docObserver = None

        pref = Preferences.preferences()
        pref.SetBool("InsertShowOnlyParts", self.form.CheckBox_ShowOnlyParts.isChecked())
        pref.SetBool("InsertRigidSubAssemblies", self.form.CheckBox_RigidSubAsm.isChecked())
        Gui.Selection.clearSelection()

    def buildPartList(self):
        self.form.partList.clear()
        self.doc_item_map.clear()

        docList = App.listDocuments().values()

        for doc in docList:
            # Create a new tree item for the document
            docItem = QtGui.QTreeWidgetItem()
            itemName = doc.Label
            icon = QIcon.fromTheme("add", QIcon(":/icons/Document.svg"))
            if doc.Partial:
                itemName = (
                    itemName + " (" + QT_TRANSLATE_NOOP("Assembly_Insert", "Partially loaded") + ")"
                )
                icon = self.createDisabledIcon(icon)
            docItem.setText(0, itemName)
            docItem.setIcon(0, icon)
            self.doc_item_map[docItem] = doc

            if not any(
                (child.isDerivedFrom("Part::Feature") or child.isDerivedFrom("App::Part"))
                for child in doc.Objects
            ):
                continue  # Skip this doc if no relevant objects

            self.form.partList.addTopLevelItem(docItem)

            def process_objects(objs, item):
                onlyParts = self.form.CheckBox_ShowOnlyParts.isChecked()
                for obj in objs:
                    if obj == self.assembly:
                        continue  # Skip current assembly

                    if obj in self.assembly.InListRecursive:
                        continue  # Prevent dependency loop.
                        # For instance if asm1/asm2 with asm2 active, we don't want to have asm1 in the list

                    if not obj.ViewObject.ShowInTree and not self.showHidden:
                        continue

                    if (
                        obj.isDerivedFrom("Part::Feature")
                        or obj.isDerivedFrom("App::Part")
                        or obj.isDerivedFrom("App::DocumentObjectGroup")
                    ):
                        # Special handling for DocumentObjectGroup: only add if it contains relevant child objects
                        if obj.isDerivedFrom("App::DocumentObjectGroup"):
                            if not any(
                                (
                                    (not onlyParts and child.isDerivedFrom("Part::Feature"))
                                    or child.isDerivedFrom("App::Part")
                                )
                                for child in obj.ViewObject.claimChildrenRecursive()
                            ):
                                continue  # Skip this object if no relevant children

                        if obj.isDerivedFrom("Part::Feature"):
                            if onlyParts:
                                continue  # Ignore solids if we show only Parts

                        # Now add the object under the document item
                        objItem = QtGui.QTreeWidgetItem(item)
                        objItem.setText(0, obj.Label)
                        objItem.setIcon(
                            0, obj.ViewObject.Icon if hasattr(obj, "ViewObject") else QtGui.QIcon()
                        )  # Use object's icon if available

                        if not obj.isDerivedFrom("App::DocumentObjectGroup"):
                            objItem.setData(0, QtCore.Qt.UserRole, obj)

                        if obj.isDerivedFrom("App::Part") or obj.isDerivedFrom(
                            "App::DocumentObjectGroup"
                        ):
                            process_objects(obj.ViewObject.claimChildren(), objItem)

            guiDoc = Gui.getDocument(doc.Name)
            process_objects(guiDoc.TreeRootObjects, docItem)
            self.form.partList.expandAll()

        self.adjustTreeWidgetSize()

    def adjustTreeWidgetSize(self):
        # Adjust the height of the part list based on item count
        item_count = 1

        def count_items(item):
            nonlocal item_count
            item_count += 1
            for i in range(item.childCount()):
                count_items(item.child(i))

        for i in range(self.form.partList.topLevelItemCount()):
            count_items(self.form.partList.topLevelItem(i))

        item_height = self.form.partList.sizeHintForRow(0)
        total_height = item_count * item_height
        max_height = 500

        self.form.partList.setMinimumHeight(min(total_height, max_height))

    def onFilterChange(self):
        filter_str = self.form.filterPartList.text().strip().lower()

        def filter_tree_item(item):
            # This function recursively filters items based on the filter string.
            item_text = item.text(0).lower()  # Assuming the relevant text is in the first column
            is_visible = filter_str in item_text if filter_str else True

            child_count = item.childCount()
            for i in range(child_count):
                child = item.child(i)
                child_is_visible = filter_tree_item(child)  # Recursively filter children
                is_visible = (
                    is_visible or child_is_visible
                )  # Parent is visible if any child matches

            item.setHidden(not is_visible)
            return is_visible

        root_count = self.form.partList.topLevelItemCount()
        for i in range(root_count):
            root_item = self.form.partList.topLevelItem(i)
            filter_tree_item(root_item)  # Filter from each root item

    def openFiles(self):
        selected_files, _ = QtGui.QFileDialog.getOpenFileNames(
            None,
            "Select FreeCAD documents to import parts from",
            "",
            "Supported Formats (*.FCStd *.fcstd);;All files (*)",
        )

        for filename in selected_files:
            requested_file = os.path.split(filename)[1]
            import_doc_is_open = any(
                requested_file == os.path.split(doc.FileName)[1]
                for doc in App.listDocuments().values()
            )

            if not import_doc_is_open:
                if filename.lower().endswith(".fcstd"):
                    App.openDocument(filename, True)
                    App.setActiveDocument(self.doc.Name)
                    self.buildPartList()

    def onItemClicked(self, item):
        selectedPart = item.data(0, QtCore.Qt.UserRole)
        if not selectedPart:
            # If there's no part associated, toggle the expanded state
            item.setExpanded(not item.isExpanded())
            return

        # check that the current document had been saved or that it's the same document as that of the selected part
        if not self.doc == selectedPart.Document:
            if self.doc.FileName == "":
                msgBox = QtWidgets.QMessageBox()
                msgBox.setIcon(QtWidgets.QMessageBox.Warning)
                msgBox.setText(
                    "The current document must be saved before inserting external parts."
                )
                msgBox.setWindowTitle("Save Document")
                saveButton = msgBox.addButton("Save", QtWidgets.QMessageBox.AcceptRole)
                cancelButton = msgBox.addButton("Cancel", QtWidgets.QMessageBox.RejectRole)

                msgBox.exec_()

                if not (msgBox.clickedButton() == saveButton and Gui.ActiveDocument.saveAs()):
                    return

            # check that the selectedPart document is saved.
            if selectedPart.Document.FileName == "":
                msgBox = QtWidgets.QMessageBox()
                msgBox.setIcon(QtWidgets.QMessageBox.Warning)
                msgBox.setText("The selected object's document must be saved before inserting it.")
                msgBox.setWindowTitle("Save Document")
                saveButton = msgBox.addButton("Save", QtWidgets.QMessageBox.AcceptRole)
                cancelButton = msgBox.addButton("Cancel", QtWidgets.QMessageBox.RejectRole)

                msgBox.exec_()

                if not (
                    msgBox.clickedButton() == saveButton
                    and selectedPart.ViewObject.Document.saveAs()
                ):
                    return

                # Update the document item text - useless because Document.Name still return 'Unnamed'
                """documentItem = item
                while documentItem.parent() is not None:
                    documentItem = documentItem.parent()
                newDocName = selectedPart.Document.Name
                print(selectedPart.Document.Name)
                documentItem.setText(0, f"{newDocName}.FCStd")"""

        if selectedPart.isDerivedFrom("Assembly::AssemblyObject"):
            objType = "Assembly::AssemblyLink"
        else:
            objType = "App::Link"

        addedObject = self.assembly.newObject(objType, selectedPart.Label)

        # set placement of the added object to the center of the screen.
        view = Gui.activeView()
        x, y = view.getSize()
        screenCenter = view.getPointOnFocalPlane(x // 2, y // 2)
        screenCorner = view.getPointOnFocalPlane(x, y)

        addedObject.LinkedObject = selectedPart
        addedObject.Label = selectedPart.Label  # non-ASCII characters fails with newObject. #12164
        addedObject.recompute()

        insertionDict = {}
        insertionDict["item"] = item
        insertionDict["addedObject"] = addedObject
        self.insertionStack.append(insertionDict)
        self.increment_counter(item)

        translation = App.Vector()
        resetThreshold = (screenCorner - screenCenter).Length * 0.1
        if len(self.insertionStack) == 1:
            translation = App.Vector()  # No translation for first object.
        elif (self.prevScreenCenter - screenCenter).Length > resetThreshold:
            self.totalTranslation = App.Vector()
            self.prevScreenCenter = screenCenter
        else:
            translation = self.getTranslationVec(addedObject)

        insertionDict["translation"] = translation
        self.totalTranslation += translation

        originX, originY = view.getPointOnViewport(App.Vector() + translation)
        if originX > 0 and originX < x and originY > 0 and originY < y:
            # If the origin is within view then we insert at the origin.
            addedObject.Placement.Base = self.totalTranslation
        else:
            #
            bboxCenter = addedObject.ViewObject.getBoundingBox().Center
            addedObject.Placement.Base = screenCenter - bboxCenter + self.totalTranslation

        self.prevScreenCenter = screenCenter

        # We turn it flexible after changing the position so that it uses the logic in
        # AssemblyLink::onChanged to handle positioning correctly.
        if selectedPart.isDerivedFrom("Assembly::AssemblyObject"):
            addedObject.Rigid = self.form.CheckBox_RigidSubAsm.isChecked()

        # highlight the link
        Gui.Selection.clearSelection()
        Gui.Selection.addSelection(self.doc.Name, addedObject.Name, "")

        item.setSelected(False)

        if len(self.insertionStack) == 1 and not UtilsAssembly.isAssemblyGrounded():
            self.handleFirstInsertion()

    def handleFirstInsertion(self):
        pref = Preferences.preferences()
        fixPart = False
        fixPartPref = pref.GetInt("GroundFirstPart", 0)
        if fixPartPref == 0:  # unset
            msgBox = QtWidgets.QMessageBox()
            msgBox.setWindowTitle("Ground Part?")
            msgBox.setText(
                "Do you want to ground the first inserted part automatically?\nYou need at least one grounded part in your assembly."
            )
            msgBox.setIcon(QtWidgets.QMessageBox.Question)

            yesButton = msgBox.addButton("Yes", QtWidgets.QMessageBox.YesRole)
            noButton = msgBox.addButton("No", QtWidgets.QMessageBox.RejectRole)
            yesAlwaysButton = msgBox.addButton("Always", QtWidgets.QMessageBox.YesRole)
            noAlwaysButton = msgBox.addButton("Never", QtWidgets.QMessageBox.NoRole)

            msgBox.exec_()

            clickedButton = msgBox.clickedButton()
            if clickedButton == yesButton:
                fixPart = True
            elif clickedButton == yesAlwaysButton:
                fixPart = True
                pref.SetInt("GroundFirstPart", 1)
            elif clickedButton == noAlwaysButton:
                pref.SetInt("GroundFirstPart", 2)

        elif fixPartPref == 1:  # Yes always
            fixPart = True

        if fixPart:
            # Create groundedJoint.
            if len(self.insertionStack) != 1:
                return

            targetObj = self.insertionStack[0]["addedObject"]

            # If the object is a flexible AssemblyLink, we should ground its internal 'base' part
            if targetObj.isDerivedFrom("Assembly::AssemblyLink") and not targetObj.Rigid:
                linkedAsm = targetObj.LinkedObject
                if linkedAsm and hasattr(linkedAsm, "Group"):
                    srcGrounded = None
                    # Attempt to find the grounded joint in the source assembly
                    # We look for a joint where JointType is 'Grounded'
                    for obj in linkedAsm.InListRecursive:
                        if hasattr(obj, "ObjectToGround"):
                            srcGrounded = obj.ObjectToGround
                            break

                    # Search the sub-assembly group for the link pointing to the source grounded object
                    # Fallback to the first valid part if no grounded joint was found in source
                    candidate = None
                    for child in targetObj.Group:
                        if not candidate and (
                            child.isDerivedFrom("App::Link") or child.isDerivedFrom("Part::Feature")
                        ):
                            candidate = child

                        if (
                            srcGrounded
                            and hasattr(child, "LinkedObject")
                            and child.LinkedObject == srcGrounded
                        ):
                            candidate = child
                            break

                    if not candidate:  # Nothing to ground
                        return

                    targetObj = candidate

            self.groundedObj = targetObj
            self.groundedJoint = CommandCreateJoint.createGroundedJoint(self.groundedObj)

    def increment_counter(self, item):
        text = item.text(0)
        match = re.search(r"(\d+) inserted$", text)

        if match:
            # Counter exists, increment it
            counter = int(match.group(1)) + 1
            new_text = re.sub(r"\d+ inserted$", f"{counter} inserted", text)
        else:
            # Counter does not exist, add it
            new_text = f"{text} : 1 inserted"

        item.setText(0, new_text)

    def decrement_counter(self, item):
        text = item.text(0)
        match = re.search(r"(\d+) inserted$", text)

        if match:
            counter = int(match.group(1)) - 1
            if counter > 0:
                # Update the counter
                new_text = re.sub(r"\d+ inserted$", f"{counter} inserted", text)
            elif counter == 0:
                # Remove the counter part from the text
                new_text = re.sub(r" : \d+ inserted$", "", text)
            else:
                return

            item.setText(0, new_text)

    """def clickMouse(self, info):
        if info["Button"] == "BUTTON1" and info["State"] == "DOWN":
            Gui.Selection.clearSelection()
            if info["ShiftDown"]:
                # Create a new link and moves this one now
                addedObject = self.insertionStack[-1]["addedObject"]
                currentPos = addedObject.Placement.Base
                selectedPart = addedObject
                if addedObject.TypeId == "App::Link":
                    selectedPart = addedObject.LinkedObject

                addedObject = self.assembly.newObject("App::Link", selectedPart.Label)
                addedObject.LinkedObject = selectedPart
                addedObject.Placement.Base = currentPos

                insertionDict = {}
                insertionDict["translation"] = App.Vector()
                insertionDict["item"] = self.insertionStack[-1]["item"]
                insertionDict["addedObject"] = addedObject
                self.insertionStack.append(insertionDict)

            else:
                self.endMove()

        elif info["Button"] == "BUTTON2" and info["State"] == "DOWN":
            self.dismissPart()"""

    # Taskbox keyboard event handler
    def eventFilter(self, watched, event):

        if event.type() == QtCore.QEvent.ContextMenu and watched is self.form.partList:
            item = watched.itemAt(event.pos())

            if item:
                if item.parent() is None:
                    doc = self.doc_item_map.get(item)
                    if doc and doc.Partial:
                        menu = QtWidgets.QMenu()
                        load_action_text = QT_TRANSLATE_NOOP(
                            "Assembly_Insert", "Fully load document"
                        )
                        load_action = menu.addAction(load_action_text)
                        load_action.triggered.connect(lambda: self.fullyLoadDocument(doc))
                        menu.exec_(event.globalPos())
                        return True  # Event was handled

                # Iterate through the insertionStack in reverse
                for i in reversed(range(len(self.insertionStack))):
                    stack_item = self.insertionStack[i]

                    if stack_item["item"] == item:
                        obj = stack_item["addedObject"]

                        # ONLY remove the object from the document.
                        # The Observer (onObjectDeleted) will handle the rest.
                        if obj and obj.Document:
                            UtilsAssembly.removeObjAndChilds(obj)

                        return True
            else:
                menu = QtWidgets.QMenu()

                # Add the checkbox action
                showHiddenAction = QtWidgets.QAction("Show objects hidden in tree view", menu)
                showHiddenAction.setCheckable(True)
                showHiddenAction.setChecked(self.showHidden)

                # Connect the action to toggle `self.showHidden`
                showHiddenAction.toggled.connect(self.toggleShowHidden)
                menu.addAction(showHiddenAction)
                menu.exec_(event.globalPos())
                return True

        return super().eventFilter(watched, event)

    def fullyLoadDocument(self, doc_to_load):
        """Closes and re-opens a document to load it fully."""
        if not doc_to_load.FileName:
            return

        # Save UI state
        scrollbar = self.form.partList.verticalScrollBar()
        scroll_position = scrollbar.value()

        # Perform the reload
        App.open(doc_to_load.FileName)
        App.setActiveDocument(self.doc.Name)

        # Refresh the UI
        self.buildPartList()

        # Restore UI state
        scrollbar.setValue(scroll_position)

    def createDisabledIcon(self, icon):
        if icon.isNull():
            return QIcon()

        # Get a pixmap from the icon at a standard size
        pixmap = icon.pixmap(icon.actualSize(QtCore.QSize(16, 16)))

        # Ask the application's style to generate a disabled version of the pixmap
        style = QtWidgets.QApplication.style()
        disabled_pixmap = style.generatedIconPixmap(
            QtGui.QIcon.Disabled, pixmap, QtWidgets.QStyleOption()
        )

        return QIcon(disabled_pixmap)

    def toggleShowHidden(self, checked):
        self.showHidden = checked
        self.buildPartList()

    def getTranslationVec(self, part):
        bb = part.Shape.BoundBox
        if bb:
            translation = (bb.XMax + bb.YMax + bb.ZMax) * 0.15
        else:
            translation = 10
        return App.Vector(translation, translation, translation)

    def onObjectDeleted(self, obj):
        """
        Handles cleanup when an object is deleted (via Right Click OR Tree View).
        """
        # Iterate backwards to safely delete
        for i in reversed(range(len(self.insertionStack))):
            stack_item = self.insertionStack[i]

            if stack_item["addedObject"] == obj:
                # 1. Revert translation
                self.totalTranslation -= stack_item["translation"]

                # 2. Update UI counter
                item = stack_item["item"]
                self.decrement_counter(item)

                # 3. Handle Grounded Joint cleanup
                if self.groundedObj == obj:
                    if self.groundedJoint:
                        try:
                            # Remove the joint if it still exists
                            if self.groundedJoint.Document:
                                self.groundedJoint.Document.removeObject(self.groundedJoint.Name)
                        except Exception:
                            pass
                    self.groundedObj = None
                    self.groundedJoint = None

                # 4. Remove from stack
                del self.insertionStack[i]

                # 5. Clear selection
                if item:
                    item.setSelected(False)


if App.GuiUp:
    Gui.addCommand("Assembly_InsertLink", CommandInsertLink())
    Gui.addCommand("Assembly_Insert", CommandGroupInsert())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Misc Arch util commands"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_CurtainWall:
    "the Arch Curtain Wall command definition"

    def GetResources(self):

        return {
            "Pixmap": "Arch_CurtainWall",
            "MenuText": QT_TRANSLATE_NOOP("Arch_CurtainWall", "Curtain Wall"),
            "Accel": "C, W",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_CurtainWall",
                "Creates a curtain wall object from selected line or from scratch",
            ),
        }

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        self.doc = FreeCAD.ActiveDocument
        sel = FreeCADGui.Selection.getSelection()
        if len(sel) > 1:
            FreeCAD.Console.PrintError(
                translate("Arch", "Select only one base object or none") + "\n"
            )
        elif len(sel) == 1:
            # build on selection
            FreeCADGui.Control.closeDialog()
            self.doc.openTransaction(translate("Arch", "Create Curtain Wall"))
            FreeCADGui.addModule("Draft")
            FreeCADGui.addModule("Arch")
            FreeCADGui.doCommand(
                "obj = Arch.makeCurtainWall(FreeCAD.ActiveDocument."
                + FreeCADGui.Selection.getSelection()[0].Name
                + ")"
            )
            FreeCADGui.doCommand("Draft.autogroup(obj)")
            self.doc.commitTransaction()
            self.doc.recompute()
        else:
            # interactive line drawing
            FreeCAD.activeDraftCommand = self  # register as a Draft command for auto grid on/off
            self.points = []
            import WorkingPlane

            WorkingPlane.get_working_plane()
            if hasattr(FreeCADGui, "Snapper"):
                FreeCADGui.Snapper.getPoint(callback=self.getPoint)

    def getPoint(self, point=None, obj=None):
        """Callback for clicks during interactive mode"""

        if point is None:
            # cancelled
            FreeCAD.activeDraftCommand = None
            FreeCADGui.Snapper.off()
            return
        self.points.append(point)
        if len(self.points) == 1:
            FreeCADGui.Snapper.getPoint(last=self.points[0], callback=self.getPoint)
        elif len(self.points) == 2:
            FreeCAD.activeDraftCommand = None
            FreeCADGui.Snapper.off()
            FreeCADGui.Control.closeDialog()
            self.doc.openTransaction(translate("Arch", "Create Curtain Wall"))
            FreeCADGui.addModule("Draft")
            FreeCADGui.addModule("Arch")
            FreeCADGui.doCommand(
                "base = Draft.makeLine(FreeCAD."
                + str(self.points[0])
                + ",FreeCAD."
                + str(self.points[1])
                + ")"
            )
            FreeCADGui.doCommand("obj = Arch.makeCurtainWall(base)")
            FreeCADGui.doCommand("Draft.autogroup(obj)")
            self.doc.commitTransaction()
            self.doc.recompute()


FreeCADGui.addCommand("Arch_CurtainWall", Arch_CurtainWall())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM EmptyTrash command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP


class BIM_EmptyTrash:

    def GetResources(self):
        return {
            "Pixmap": "BIM_Trash",
            "MenuText": QT_TRANSLATE_NOOP("BIM_EmptyTrash", "Empty Trash"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_EmptyTrash",
                "Deletes all objects from the trash bin that are not used by any other",
            ),
        }

    def Activated(self):
        trash = FreeCAD.ActiveDocument.getObject("Trash")
        if trash and trash.isDerivedFrom("App::DocumentObjectGroup"):
            deletelist = []
            for obj in trash.Group:
                if (len(obj.InList) == 1) and (obj.InList[0] == trash):
                    deletelist.append(obj.Name)
                    deletelist.extend(self.getDeletableChildren(obj))
            if deletelist:
                FreeCAD.ActiveDocument.openTransaction("Empty Trash")
                for name in deletelist:
                    FreeCAD.ActiveDocument.removeObject(name)
                FreeCAD.ActiveDocument.commitTransaction()

    def getDeletableChildren(self, obj):
        deletelist = []
        for child in obj.OutList:
            if (len(child.InList) == 1) and (child.InList[0] == obj):
                deletelist.append(child.Name)
                deletelist.extend(self.getDeletableChildren(child))
        return deletelist


FreeCADGui.addCommand("BIM_EmptyTrash", BIM_EmptyTrash())



# SPDX-License-Identifier: LGPL-2.1-or-later

# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import yaml
from typing import List, Optional, Mapping, Type
from ...assets.serializer import AssetSerializer
from ...assets.uri import AssetUri
from ...shape import ToolBitShape
from ..models.base import ToolBit


class YamlToolBitSerializer(AssetSerializer):
    """
    Serializes and deserializes ToolBit instances to and from YAML.
    """

    for_class: Type[ToolBit] = ToolBit
    extensions: tuple[str, ...] = (".yaml", ".yml")
    mime_type: str = "application/x-yaml"
    can_import: bool = True
    can_export: bool = True

    @classmethod
    def get_label(cls) -> str:
        return "YAML ToolBit"

    @classmethod
    def extract_dependencies(cls, data: bytes) -> List[AssetUri]:
        """Extracts URIs of dependencies from serialized data."""
        data_dict = yaml.safe_load(data)
        if isinstance(data_dict, dict):
            shape_id = data_dict.get("shape")
            if shape_id:
                # Assuming shape is identified by its ID/name
                return [ToolBitShape.resolve_name(str(shape_id))]
        return []

    @classmethod
    def serialize(cls, asset: ToolBit) -> bytes:
        """Serializes a ToolBit instance to bytes (shallow)."""
        # Shallow serialization: only serialize direct attributes and shape ID
        data = asset.to_dict()
        return yaml.dump(data, default_flow_style=False).encode("utf-8")

    @classmethod
    def deserialize(
        cls,
        data: bytes,
        id: str | None = None,
        dependencies: Optional[Mapping[AssetUri, ToolBitShape]] = None,
    ) -> ToolBit:
        """
        Creates a ToolBit instance from serialized data and resolved
        dependencies (shallow).
        """
        data_dict = yaml.safe_load(data)
        if not isinstance(data_dict, dict):
            raise ValueError("Invalid YAML data for ToolBit")
        toolbit = ToolBit.from_dict(data_dict)
        if id:
            toolbit.id = id
        return toolbit

    @classmethod
    def deep_deserialize(cls, data: bytes) -> ToolBit:
        """Deep deserialize preserving the original toolbit ID."""
        data_dict = yaml.safe_load(data)
        if not isinstance(data_dict, dict):
            raise ValueError("Invalid YAML data for ToolBit")

        original_id = data_dict.get("id")  # Extract the original ID
        toolbit = ToolBit.from_dict(data_dict)
        if original_id:
            toolbit.id = original_id  # Preserve the original ID
        return toolbit



# ***************************************************************************
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint section print task panel for the document object"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package task_constraint_sectionprint
#  \ingroup FEM
#  \brief task panel for constraint section print object

from PySide import QtCore
from PySide import QtGui

import FreeCAD
import FreeCADGui

from femguiutils import selection_widgets
from . import base_femtaskpanel


# TODO uses the old style Add button, move to the new style. See constraint fixed
class _TaskPanel(base_femtaskpanel._BaseTaskPanel):
    """
    The TaskPanel for editing References property of ConstraintSectionPrint objects
    """

    def __init__(self, obj):
        super().__init__(obj)

        # parameter widget
        self.parameterWidget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/ConstraintSectionPrint.ui"
        )

        self.init_parameter_widget()

        QtCore.QObject.connect(
            self.parameterWidget.cb_variable,
            QtCore.SIGNAL("currentIndexChanged(int)"),
            self.variable_changed,
        )

        # geometry selection widget
        self.selectionWidget = selection_widgets.GeometryElementsSelection(
            obj.References, ["Face"], False, False
        )

        # form made from param and selection widget
        self.form = [self.selectionWidget, self.parameterWidget]

    def accept(self):
        # check values
        items = len(self.selectionWidget.references)

        if items != 1:
            msgBox = QtGui.QMessageBox()
            msgBox.setIcon(QtGui.QMessageBox.Question)
            msgBox.setText(
                "Constraint SectionPrint requires exactly one face\n\nfound references: {}".format(
                    items
                )
            )
            msgBox.setWindowTitle("FreeCAD FEM Constraint SectionPrint")
            retryButton = msgBox.addButton(QtGui.QMessageBox.Retry)
            ignoreButton = msgBox.addButton(QtGui.QMessageBox.Ignore)
            msgBox.exec_()

            if msgBox.clickedButton() == retryButton:
                return False
            elif msgBox.clickedButton() == ignoreButton:
                pass

        self.obj.Variable = self.variable
        self.obj.References = self.selectionWidget.references
        self.selectionWidget.finish_selection()
        return super().accept()

    def reject(self):
        self.selectionWidget.finish_selection()
        return super().reject()

    def init_parameter_widget(self):
        self.variable = self.obj.Variable
        self.variable_enum = self.obj.getEnumerationsOfProperty("Variable")
        self.parameterWidget.cb_variable.addItems(self.variable_enum)
        index = self.variable_enum.index(self.variable)
        self.parameterWidget.cb_variable.setCurrentIndex(index)

    def variable_changed(self, index):
        self.variable = self.variable_enum[index]
